# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

r"""Class for the estimation of Sobol' indices.

Let us consider the model :math:`Y=f(X_1,\ldots,X_d)`
where:

- :math:`X_1,\ldots,X_d` are independent random variables,
- :math:`E\left[f(X_1,\ldots,X_d)^2\right]<\infty`.

Then, the following decomposition is unique:

.. math::

   Y=f_0 + \sum_{i=1}^df_i(X_i) + \sum_{i,j=1\atop i\neq j}^d f_{i,j}(X_i,X_j)
   + \sum_{i,j,k=1\atop i\neq j\neq k}^d f_{i,j,k}(X_i,X_j,X_k) + \ldots +
   f_{1,\ldots,d}(X_1,\ldots,X_d)

where:

- :math:`f_0=E[Y]`,
- :math:`f_i(X_i)=E[Y|X_i]-f_0`,
- :math:`f_{i,j}(X_i,X_j)=E[Y|X_i,X_j]-f_i(X_i)-f_j(X_j)-f_0`
- and so on.

Then, the shift to variance leads to:

.. math::

   V[Y]=\sum_{i=1}^dV\left[f_i(X_i)\right] +
   \sum_{i,j=1\atop j\neq i}^d V\left[f_{i,j}(X_i,X_j)\right] + \ldots +
   V\left[f_{1,\ldots,d}(X_1,\ldots,X_d)\right]

and the Sobol' indices are obtained by dividing by the variance and sum up to 1:

.. math::

   1=\sum_{i=1}^dS_i + \sum_{i,j=1\atop j\neq i}^d S_{i,j} +
   \sum_{i,j,k=1\atop i\neq j\neq k}^d S_{i,j,k} + \ldots + S_{1,\ldots,d}

A Sobol' index represents the share of output variance explained
by a parameter or a group of parameters. For the parameter :math:`X_i`,

- :math:`S_i` is the first-order Sobol' index
  measuring the individual effect of :math:`X_i`,
- :math:`S_{i,j}` is the second-order Sobol' index
  measuring the joint effect between :math:`X_i` and :math:`X_j`,
- :math:`S_{i,j,k}` is the third-order Sobol' index
  measuring the joint effect between :math:`X_i`, :math:`X_j` and :math:`X_k`,
- and so on.

In practice, we only consider the first-order Sobol' index:

.. math::

   S_i=\frac{V[E[Y|X_i]]}{V[Y]}

and the total-order Sobol' index:

.. math::

   S_i^T=\sum_{u\subset\{1,\ldots,d\}\atop u \ni i}S_u

The latter represents the sum of the individual effect of :math:`X_i` and
the joint effects between :math:`X_i` and any parameter or group of parameters.

This methodology relies on the :class:`.SobolAnalysis` class. Precisely,
:attr:`.SobolAnalysis.indices` contains
both :attr:`.SobolAnalysis.first_order_indices` and
:attr:`.SobolAnalysis.total_order_indices`
while :attr:`.SobolAnalysis.main_indices` represents total-order Sobol'
indices.
Lastly, the :meth:`.SobolAnalysis.plot` method represents
the estimations of both first-order and total-order Sobol' indices along with
their 95% confidence interval.

The user can select the algorithm to estimate the Sobol' indices.
The computation relies on
`OpenTURNS capabilities <http://www.openturns.org/>`_.
"""

from __future__ import division, unicode_literals

import logging
from typing import Dict, Iterable, Mapping, Optional, Sequence, Tuple, Union

import matplotlib.pyplot as plt
from matplotlib.transforms import Affine2D
from numpy import array
from openturns import (
    JansenSensitivityAlgorithm,
    MartinezSensitivityAlgorithm,
    MauntzKucherenkoSensitivityAlgorithm,
    SaltelliSensitivityAlgorithm,
    Sample,
)
from past.utils import old_div

from gemseo.algos.doe.lib_openturns import OpenTURNS
from gemseo.algos.parameter_space import ParameterSpace
from gemseo.core.discipline import MDODiscipline
from gemseo.uncertainty.sensitivity.analysis import IndicesType, SensitivityAnalysis
from gemseo.utils.data_conversion import DataConversion
from gemseo.utils.py23_compat import Path

LOGGER = logging.getLogger(__name__)


class SobolAnalysis(SensitivityAnalysis):
    """Sensitivity analysis based on the Sobol' indices.

    Examples:
        >>> from numpy import pi
        >>> from gemseo.api import create_discipline, create_parameter_space
        >>> from gemseo.uncertainty.sensitivity.sobol.analysis import SobolAnalysis
        >>>
        >>> expressions = {"y": "sin(x1)+7*sin(x2)**2+0.1*x3**4*sin(x1)"}
        >>> discipline = create_discipline(
        ...     "AnalyticDiscipline", expressions_dict=expressions
        ... )
        >>>
        >>> parameter_space = create_parameter_space()
        >>> parameter_space.add_random_variable(
        ...     "x1", "OTUniformDistribution", minimum=-pi, maximum=pi
        ... )
        >>> parameter_space.add_random_variable(
        ...     "x2", "OTUniformDistribution", minimum=-pi, maximum=pi
        ... )
        >>> parameter_space.add_random_variable(
        ...     "x3", "OTUniformDistribution", minimum=-pi, maximum=pi
        ... )
        >>>
        >>> analysis = SobolAnalysis(discipline, parameter_space, n_samples=10000)
        >>> indices = analysis.compute_indices()
    """

    _ALGOS = {
        "Saltelli": SaltelliSensitivityAlgorithm,
        "Jansen": JansenSensitivityAlgorithm,
        "MauntzKucherenko": MauntzKucherenkoSensitivityAlgorithm,
        "Martinez": MartinezSensitivityAlgorithm,
    }
    _FIRST = "first"
    _TOTAL = "total"
    _FIRST_METHOD = "Sobol({})".format(_FIRST)
    _TOTAL_METHOD = "Sobol({})".format(_TOTAL)

    AVAILABLE_ALGOS = sorted(_ALGOS.keys())
    DEFAULT_DRIVER = OpenTURNS.OT_SOBOL_INDICES

    def __init__(
        self,
        discipline,  # type: MDODiscipline
        parameter_space,  # type: ParameterSpace
        n_samples,  # type: int
        algo=None,  # type:Optional[str]
        algo_options=None,  # type: Optional[Mapping]
    ):  # type: (...) -> None  # noqa: D107,D205,D212,D415
        self.__sobol = None
        super(SobolAnalysis, self).__init__(discipline, parameter_space, n_samples)
        self.main_method = self._FIRST

    @SensitivityAnalysis.main_method.setter
    def main_method(
        self,
        name,  # type: str
    ):  # type:(...) -> None # noqa: D102
        if name == self._FIRST:
            self._main_method = self._FIRST_METHOD
            LOGGER.info("Use first order indices as main indices.")
        elif name == self._TOTAL:
            self._main_method = self._TOTAL_METHOD
            LOGGER.info("Use total order indices as main indices.")
        else:
            raise NotImplementedError(
                "{} is a bad method name. "
                "Available ones are {}.".format(name, [self._FIRST, self._TOTAL])
            )

    def compute_indices(
        self,
        outputs=None,  # type: Optional[Sequence[str]]
        algo="Saltelli",  # type: str
    ):  # type: (...) -> Dict[str,IndicesType]
        # noqa:D205,D212,D415
        """
        Args:
            algo: The name of the algorithm to estimate the Sobol' indices
        """
        try:
            algo = self._ALGOS[algo]
        except Exception:
            raise TypeError(
                "{} is not an available algorithm "
                "to compute Sobol indices.".format(algo)
            )
        output_names = outputs or self.default_output
        if not isinstance(output_names, list):
            output_names = [output_names]
        inputs = Sample(self.dataset.get_data_by_group(self.dataset.INPUT_GROUP))
        outputs = self.dataset.get_data_by_names(output_names, True)
        dim = self.dataset.dimension[self.dataset.INPUT_GROUP]
        n_samples = int(old_div(len(self.dataset), (dim + 2)))
        self.__sobol = {}
        for output_name, value in outputs.items():
            self.__sobol[output_name] = []
            for index in range(value.shape[1]):
                sub_outputs = Sample(value[:, index][:, None])
                self.__sobol[output_name].append(algo(inputs, sub_outputs, n_samples))
        return self.indices

    def __get_indices(
        self,
        first_order=True,  # type: bool
    ):  # type: (...) -> IndicesType
        """Get the indices, either first-order or total order.

        Args:
            first_order: The type of indices.
                If True, return first-order Sobol' indices.
                Otherwise, return total-order Sobol' indices.

        Returns:
            The Sobol' indices, either first-order or total order.
        """
        if first_order:
            method = "getFirstOrderIndices"
        else:
            method = "getTotalOrderIndices"
        array_to_dict = DataConversion.array_to_dict
        inputs_names = self.dataset.get_names(self.dataset.INPUT_GROUP)
        sizes = self.dataset.sizes
        indices = {}
        for name in self.__sobol:
            indices[name] = [
                array_to_dict(array(getattr(sobol, method)()), inputs_names, sizes)
                for sobol in self.__sobol[name]
            ]
        return indices

    @property
    def first_order_indices(self):  # type: (...) -> IndicesType
        """dict: The first-order Sobol' indices.

        With the following structure:

        .. code-block:: python

            {
                "output_name": [
                    {
                        "input_name": data_array,
                    }
                ]
            }
        """
        return self.__get_indices()

    @property
    def total_order_indices(self):  # type: (...) -> IndicesType
        """dict: The total-order Sobol' indices.

        With the following structure:

        .. code-block:: python

            {
                "output_name": [
                    {
                        "input_name": data_array,
                    }
                ]
            }
        """
        return self.__get_indices(False)

    def get_intervals(
        self,
        first_order=True,  # type: bool
    ):  # type: (...) -> IndicesType
        """Get the confidence interval for Sobol' indices.

        Args:
            first_order: The type of indices.
                If True, returns the intervals for the first-order indices.
                Otherwise, for the total-order indices.

        Returns:
            dict: The confidence intervals for the Sobol' indices.

            With the following structure:

            .. code-block:: python

                {
                    "output_name": [
                        {
                            "input_name": data_array,
                        }
                    ]
                }
        """
        array_to_dict = DataConversion.array_to_dict
        inputs_names = self.dataset.get_names(self.dataset.INPUT_GROUP)
        sizes = self.dataset.sizes
        intervals = {}
        for output_name in self.__sobol:
            intervals[output_name] = []
            for sobol in self.__sobol[output_name]:
                if first_order:
                    interval = sobol.getFirstOrderIndicesInterval()
                else:
                    interval = sobol.getTotalOrderIndicesInterval()
                lower_bound = array(interval.getLowerBound())
                upper_bound = array(interval.getUpperBound())
                lower_bound = array_to_dict(lower_bound, inputs_names, sizes)
                upper_bound = array_to_dict(upper_bound, inputs_names, sizes)
                intervals[output_name].append(
                    {
                        name: array([lower_bound[name][0], upper_bound[name][0]])
                        for name in lower_bound
                    }
                )
        return intervals

    @property
    def indices(
        self,
    ):  # type: (...) -> Dict[str, IndicesType] # noqa: D102
        return {"first": self.first_order_indices, "total": self.total_order_indices}

    @property
    def main_indices(self):  # type: (...) -> IndicesType # noqa: D102
        if self.main_method == self._TOTAL_METHOD:
            return self.total_order_indices
        else:
            return self.first_order_indices

    def plot(
        self,
        output,  # type: Union[str,Tuple[str,int]]
        inputs=None,  # type: Optional[Iterable[str]]
        title=None,  # type: Optional[str]
        save=True,  # type: bool
        show=False,  # type: bool
        file_path=None,  # type: Optional[Union[str,Path]]
        directory_path=None,  # type: Optional[Union[str,Path]]
        file_name=None,  # type: Optional[str]
        file_format=None,  # type: Optional[str]
        sort=True,  # type:bool
        sort_by_total=True,  # type:bool
    ):  # noqa: D417
        r"""Plot the first- and total-order Sobol' indices.

        For :math:`i\in\{1,\ldots,d\}`, plot :math:`S_i^{1}` and :math:`S_T^{1}`
        with their confidence intervals,

        Args:
            sort: The sorting option.
                If True, sort variables before display.
            sort_by_total: The type of sorting.
                If True, sort variables according to total-order Sobol' indices.
                Otherwise, use first-order Sobol' indices.
        """
        if not isinstance(output, tuple):
            output = (output, 0)
        fig, ax = plt.subplots()
        if sort_by_total:
            indices = self.total_order_indices
        else:
            indices = self.first_order_indices
        intervals = self.get_intervals()
        indices = indices[output[0]][output[1]]
        intervals = intervals[output[0]][output[1]]
        first_order_indices = self.first_order_indices[output[0]][output[1]]
        total_order_indices = self.total_order_indices[output[0]][output[1]]
        if sort:
            names = [
                name
                for name, _ in sorted(
                    list(indices.items()), key=lambda item: item[1], reverse=True
                )
            ]
        else:
            names = list(indices.keys())
        names = self._filter_names(names, inputs)
        errorbar_options = {"marker": "o", "linestyle": "", "markersize": 7}
        trans1 = Affine2D().translate(-0.01, 0.0) + ax.transData
        trans2 = Affine2D().translate(+0.01, 0.0) + ax.transData
        values = [first_order_indices[name] for name in names]
        yerr = array(
            [
                [
                    first_order_indices[name][0] - intervals[name][0],
                    intervals[name][1] - first_order_indices[name][0],
                ]
                for name in names
            ]
        ).T
        ax.errorbar(
            names,
            values,
            yerr,
            label="First order",
            transform=trans2,
            **errorbar_options
        )
        intervals = self.get_intervals(False)
        intervals = intervals[output[0]][output[1]]
        values = [total_order_indices[name] for name in names]
        yerr = array(
            [
                [
                    total_order_indices[name][0] - intervals[name][0],
                    intervals[name][1] - total_order_indices[name][0],
                ]
                for name in names
            ]
        ).T
        ax.errorbar(
            names,
            values,
            yerr,
            label="Total order",
            transform=trans1,
            **errorbar_options
        )
        ax.legend(
            loc="lower left",
            bbox_to_anchor=(0.0, 1.01),
            ncol=2,
            borderaxespad=0,
            frameon=False,
        )
        output = "{}({})".format(output[0], output[1])
        ax.set_title(title or "Sobol indices for the output {}".format(output))
        self._save_show_plot(
            fig,
            save=save,
            show=show,
            file_path=file_path,
            file_name=file_name,
            file_format=file_format,
            directory_path=directory_path,
        )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Dummy linear discipline."""

from typing import Optional, Sequence

from numpy import ones
from numpy.random import rand

from gemseo.core.discipline import MDODiscipline
from gemseo.utils.data_conversion import DataConversion


class LinearDiscipline(MDODiscipline):
    """A discipline that computes random outputs from inputs.

    The output are computed by a product with a random matrix and the inputs. The inputs
    and output names are specified by the user. The size of inputs and outputs can be
    specified.
    """

    def __init__(
        self,
        name,  # type: str
        input_names,  # type: Sequence[str]
        output_names,  # type: Sequence[str]
        inputs_size=1,  # type: int
        outputs_size=1,  # type: int
        grammar_type=MDODiscipline.JSON_GRAMMAR_TYPE,  # type: str
    ):  # type: (...) -> None
        """# noqa: D205,D212,D415
        Args:
            name: The discipline name.
            input_names: The input data names
            output_names: The output data names.
            inputs_size: The size of input data vectors,
                each input data is of shape (inputs_size,).
            outputs_size: The size of output data vectors,
                each output data is of shape (outputs_size,).
            grammar_type: The type of grammars.
        """
        super(LinearDiscipline, self).__init__(name, grammar_type=grammar_type)
        self.input_names = input_names
        self.output_names = output_names

        self.input_grammar.initialize_from_data_names(input_names)
        self.output_grammar.initialize_from_data_names(output_names)

        self.size_in = len(input_names) * inputs_size
        self.size_out = len(output_names) * outputs_size

        self.inputs_size = inputs_size
        self.outputs_size = outputs_size
        self.mat = rand(self.size_out, self.size_in) / self.size_in

        self.__sizes_d = {k: self.inputs_size for k in self.input_names}
        self.__sizes_d.update({k: self.outputs_size for k in self.output_names})

        self.default_inputs = {k: 0.5 * ones(inputs_size) for k in input_names}

    def _run(self):  # type: (...) -> None
        in_array = DataConversion.dict_to_array(self.local_data, self.input_names)
        out_array = self.mat.dot(in_array)
        out_dict = DataConversion.array_to_dict(
            out_array,
            self.output_names,
            self.__sizes_d,
        )
        self.local_data.update(out_dict)

    def _compute_jacobian(
        self,
        inputs=None,  # type: Optional[Sequence[str]]
        outputs=None,  # type: Optional[Sequence[str]]
    ):  # type: (...) -> None
        self.jac = DataConversion.jac_2dmat_to_dict(
            self.mat, self.output_names, self.input_names, self.__sizes_d
        )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#       :author : Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
GGOBI : interactive data visualization software
***********************************************

Export data to the XML file format needed by GGOBI
"""
from __future__ import division, unicode_literals

import os
from xml.dom import minidom
from xml.etree import ElementTree
from xml.etree.ElementTree import Comment, Element, SubElement


def prettify(elem):
    """Return a pretty-printed XML string for the Element.

    :param elem: the xml element
    """
    rough_string = ElementTree.tostring(elem, "utf-8")
    reparsed = minidom.parseString(rough_string)
    return reparsed.toprettyxml(indent="\t")


def save_data_arrays_to_xml(variables_names, values_array, file_path="opt_hist.xml"):
    """Saves an optimization history in numpy format to an xml file to be read by ggobi.

    :param variables_names: list of the variables names
    :type variables_names: list(str)
    :param values_array: the variables history (nb variables,nb iterations)
    :param file_path: the file path of the generated xml file
    :type file_path: str
    """

    if os.path.exists(file_path):
        os.remove(file_path)
    nb_records = values_array.shape[0]
    nb_variables = len(variables_names)

    root = Element("ggobidata")
    comment = Comment('DOCTYPE ggobidata SYSTEM "ggobi.dtd"')
    root.append(comment)
    ggobidata = SubElement(root, "ggobidata")
    SubElement(ggobidata, "brush", attrib={"colo": "6", "glyph": "fc 3"})
    data = SubElement(ggobidata, "data", attrib={"name": "opt_history"})
    SubElement(data, "description", attrib={"source": "Optimization history"})
    variables = SubElement(data, "variables", attrib={"count": str(nb_variables)})
    for var in variables_names:
        variables_attr = {"name": str(var)}
        SubElement(variables, "realvariable", attrib=variables_attr)
    records = SubElement(
        data,
        "records",
        attrib={"color": str(2), "count": str(nb_records), "missingValue": "NA"},
    )
    for i_rec in range(nb_records):
        rec = SubElement(
            records, "record", attrib={"color": str(1), "label": "Iter_" + str(i_rec)}
        )
        rec.text = str(values_array[i_rec, :]).replace("[", "").replace("]", "")

    with open(file_path, "w") as xml_file:
        xml_file.write(prettify(root))



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

r"""Abstract class defining the concept of probability distribution.

Overview
--------

The abstract :class:`.Distribution` class implements the concept of
`probability distribution <https://en.wikipedia.org/wiki/Probability_distribution>`_,
which is a mathematical function giving the probabilities of occurrence
of different possible outcomes of a random variable for an experiment.
The `normal distribution <https://en.wikipedia.org/wiki/Normal_distribution>`_
with its famous *bell curve* is a well-known example of probability distribution.

.. seealso::

    This abstract class is enriched by concrete ones,
    such as :class:`.OTDistribution` interfacing the OpenTURNS probability distributions
    and :class:`.SPDistribution` interfacing the SciPy probability distributions.

Construction
------------

The :class:`.Distribution` of a given uncertain variable is built
from a recognized distribution name (e.g. 'Normal' for OpenTURNS or 'norm' for SciPy),
a variable dimension, a set of parameters
and optionally a standard representation of these parameters.

Capabilities
------------

From a :class:`.Distribution`, we can easily get statistics,
such as :attr:`.Distribution.mean`,
:attr:`.Distribution.standard_deviation`. We can also get the
numerical :attr:`.Distribution.range` and
mathematical :attr:`.Distribution.support`.

.. note::

    We call mathematical *support* the set of values that the random variable
    can take in theory, e.g. :math:`]-\infty,+\infty[` for a Gaussian variable,
    and numerical *range* the set of values that it can can take in practice,
    taking into account the values rounded to zero double precision.
    Both support and range are described in terms of lower and upper bounds

We can also evaluate the cumulative density function
(:meth:`.Distribution.compute_cdf`)
for the different marginals of the random variable,
as well as the inverse cumulative density function
(:meth:`.Distribution.compute_inverse_cdf`). We can plot them,
either for a given marginal (:meth:`.Distribution.plot`)
or for all marginals (:meth:`.Distribution.plot_all`).

Lastly, we can compute realizations of the random variable
by means of the :meth:`.Distribution.compute_samples` method.
"""

from __future__ import division, unicode_literals

import logging
from typing import Callable, Iterable, List, Mapping, Optional, Tuple, Union

import matplotlib.pyplot as plt
import six
from custom_inherit import DocInheritMeta
from matplotlib.figure import Figure
from numpy import arange, array, ndarray
from past.utils import old_div

from gemseo.utils.file_path_manager import FilePathManager, FileType
from gemseo.utils.matplotlib_figure import save_show_figure
from gemseo.utils.py23_compat import Path
from gemseo.utils.string_tools import MultiLineString, pretty_repr

LOGGER = logging.getLogger(__name__)

StandardParametersType = Mapping[str, Union[str, int, float]]
ParametersType = Union[Tuple[str, int, float], StandardParametersType]


@six.add_metaclass(DocInheritMeta(abstract_base_class=True))
class Distribution(object):
    """Probability distribution related to a random variable.

    The dimension of the random variable can be greater than 1. In this case,
    the same distribution is applied to all components of the random variable
    under the hypothesis that these components are stochastically independent.

    The string representation of a distribution
    interfacing a distribution called :code:`'MyDistribution'`
    with parameters :code:`(2,3)` is 'MyDistribution(2, 3)`
    if no standard parameters are passed.
    If the standard parameters are :code:`{a: 2, b: 3}`
    (resp. :code:`{a_inv: 2, b: 3}`),
    then the standard representation is: 'MyDistribution(a=2, b=3)`
    (resp. 'MyDistribution(a_inv=0.5, b=3)`)
    Standard parameters are useful to redefine the name of the parameters.
    For example, some exponential distributions consider the notion of rate
    while other ones consider the notion of scale, which is the inverse of the rate...
    even in the background, the distribution is the same!

    Attributes:
        math_lower_bound (ndarray): The mathematical lower bound
            of the random variable.
        math_upper_bound (ndarray): The mathematical upper bound
            of the random variable.
        num_lower_bound (ndarray): The numerical lower bound
            of the random variable.
        num_upper_bound (ndarray): The numerical upper bound
            of the random variable.
        distribution (InterfacedDistributionClass): The probability distribution
            of the random variable.
        marginals (list(InterfacedDistributionClass)): The marginal distributions
            of the components of the random variable.
        dimension (int): The number of dimensions of the random variable.
        variable_name (str): The name of the random variable.
        distribution_name (str): The name of the probability distribution.
        transformation (str): The transformation applied to the random variable,
            e.g. 'sin(x)'.
        parameters (tuple or dict): The parameters of the probability distribution.
        standard_parameters (dict, optional): The standard representation
            of the parameters of the distribution, used for its string representation.
    """

    _MU = "mu"
    _SIGMA = "sigma"
    _LOWER = "lower"
    _UPPER = "upper"
    _MODE = "mode"
    _RATE = "rate"
    _LOC = "loc"

    _COMPOSED_DISTRIBUTION = None

    def __init__(
        self,
        variable,  # type: str
        interfaced_distribution,  # type: str
        parameters,  # type: ParametersType
        dimension=1,  # type: int
        standard_parameters=None,  # type: Optional[StandardParametersType]
    ):  # noqa: D205,D212,D415
        # type: (...) -> None
        """
        Args:
            variable: The name of the random variable.
            interfaced_distribution: The name of the probability distribution,
                typically the name of a class wrapped from an external library,
                such as 'Normal' for OpenTURNS or 'norm' for SciPy.
            parameters: The parameters of the class
                related to distribution.
            dimension: The dimension of the random variable.
            standard_parameters: The standard representation
                of the parameters of the probability distribution.
        """
        self.math_lower_bound = None
        self.math_upper_bound = None
        self.num_lower_bound = None
        self.num_upper_bound = None
        self.distribution = None
        self.marginals = None
        self.dimension = dimension
        self.variable_name = variable
        self.distribution_name = interfaced_distribution
        self.transformation = variable
        self.parameters = parameters
        if standard_parameters is None:
            self.standard_parameters = self.parameters
        else:
            self.standard_parameters = standard_parameters
        self.__file_path_manager = FilePathManager(
            FileType.FIGURE, default_name="distribution_{}".format(self.variable_name)
        )
        msg = MultiLineString()
        msg.add("Define the random variable: {}", variable)
        msg.indent()
        msg.add("Distribution: {}", self)
        msg.add("Dimension: {}", dimension)
        LOGGER.info("%s", msg)

    def __str__(self):
        # type: (...) -> str
        parameters = pretty_repr(self.standard_parameters)
        return "{}({})".format(self.distribution_name, parameters)

    def compute_samples(
        self,
        n_samples=1,  # type: int
    ):
        # type: (...) -> ndarray
        """Sample the random variable.

        Args:
            n_samples: The number of samples.

        Returns:
            The samples of the random variable,

            The number of columns is equal to the dimension of the variable
            and the number of lines is equal to the number of samples.
        """
        raise NotImplementedError

    def compute_cdf(
        self,
        vector,  # type: Iterable[float]
    ):
        # type: (...) -> ndarray
        """Evaluate the cumulative density function (CDF).

        Evaluate the CDF of the components of the random variable
        for a given realization of this random variable.

        Args:
            vector: A realization of the random variable.

        Returns:
            The CDF values of the components of the random variable.
        """
        raise NotImplementedError

    def compute_inverse_cdf(
        self,
        vector,  # type: Iterable[float]
    ):
        # type: (...) -> ndarray
        """Evaluate the inverse of the cumulative density function (ICDF).

        Args:
            vector: A vector of values comprised between 0 and 1
                whose length is equal to the dimension of the random variable.

        Returns:
            The ICDF values of the components of the random variable.
        """
        raise NotImplementedError

    @property
    def mean(self):
        # type: (...) -> ndarray
        """The analytical mean of the random variable."""
        raise NotImplementedError

    @property
    def standard_deviation(self):
        # type: (...) -> ndarray
        """The analytical standard deviation of the random variable."""
        raise NotImplementedError

    @property
    def range(self):
        # type: (...) -> List[ndarray]
        """The numerical range.

        The numerical range is the interval defined by
        the lower and upper bounds numerically reachable by the random variable.

        Here, the numerical range of the random variable is defined
        by one array for each component of the random variable,
        whose first element is the lower bound of this component
        while the second one is its upper bound.
        """
        value = [
            array([l_b, u_b])
            for l_b, u_b in zip(self.num_lower_bound, self.num_upper_bound)
        ]
        return value

    @property
    def support(self):
        # type: (...) -> List[ndarray]
        """The mathematical support.

        The mathematical support is the interval defined by
        the theoretical lower and upper bounds of the random variable.

        Here, the mathematical range of the random variable is defined
        by one array for each component of the random variable,
        whose first element is the lower bound of this component
        while the second one is its upper bound.
        """
        value = [
            array([l_b, u_b])
            for l_b, u_b in zip(self.math_lower_bound, self.math_upper_bound)
        ]
        return value

    def plot_all(
        self,
        show=True,  # type: bool
        save=False,  # type: bool
        file_path=None,  # type: Optional[Union[str,Path]]
        directory_path=None,  # type: Optional[Union[str,Path]]
        file_name=None,  # type: Optional[str]
        file_extension=None,  # type: Optional[str]
    ):
        # type: (...) -> List[Figure]
        """Plot both probability and cumulative density functions for all components.

        Args:
            save: If True, save the figure.
            show: If True, display the figure.
            file_path: The path of the file to save the figures.
                If the extension is missing, use ``file_extension``.
                If None,
                create a file path
                from ``directory_path``, ``file_name`` and ``file_extension``.
            directory_path: The path of the directory to save the figures.
                If None, use the current working directory.
            file_name: The name of the file to save the figures.
                If None, use a default one generated by the post-processing.
            file_extension: A file extension, e.g. 'png', 'pdf', 'svg', ...
                If None, use a default file extension.

        Returns:
            The figures.
        """
        figures = []
        for index in range(self.dimension):
            figures.append(
                self.plot(
                    index=index,
                    show=show,
                    save=save,
                    file_path=file_path,
                    file_name=file_name,
                    file_extension=file_extension,
                    directory_path=directory_path,
                )
            )
        return figures

    def plot(
        self,
        index=0,  # type: int
        show=True,  # type: bool
        save=False,  # type: bool
        file_path=None,  # type: Optional[Union[str,Path]]
        directory_path=None,  # type: Optional[Union[str,Path]]
        file_name=None,  # type: Optional[str]
        file_extension=None,  # type: Optional[str]
    ):
        # type: (...) -> Figure
        """Plot both probability and cumulative density functions for a given component.

        Args:
            index: The index of a component of the random variable.
            save: If True, save the figure.
            show: If True, display the figure.
            file_path: The path of the file to save the figures.
                If the extension is missing, use ``file_extension``.
                If None,
                create a file path
                from ``directory_path``, ``file_name`` and ``file_extension``.
            directory_path: The path of the directory to save the figures.
                If None, use the current working directory.
            file_name: The name of the file to save the figures.
                If None, use a default one generated by the post-processing.
            file_extension: A file extension, e.g. 'png', 'pdf', 'svg', ...
                If None, use a default file extension.

        Returns:
            The figure.
        """
        variable_name = self.variable_name
        if self.dimension > 1:
            variable_name = "{}({})".format(variable_name, index)
        l_b = self.num_lower_bound[index]
        u_b = self.num_upper_bound[index]
        x_values = arange(l_b, u_b, old_div((u_b - l_b), 100))
        y1_values = [self._pdf(index)(x_value) for x_value in x_values]
        fig, (ax1, ax2) = plt.subplots(1, 2)
        fig.suptitle("Probability distribution of {}".format(variable_name))
        ax1.plot(x_values, y1_values)
        ax1.set_xlabel(variable_name)
        ax1.set_title("PDF")
        y2_values = [self._cdf(index)(x_value) for x_value in x_values]
        ax2.plot(x_values, y2_values)
        ax2.set_xlabel(variable_name)
        ax2.set_title("Cumulative density function")
        if save:
            file_path = self.__file_path_manager.create_file_path(
                file_path=file_path,
                file_name=file_name,
                directory_path=directory_path,
                file_extension=file_extension,
            )
            if self.dimension > 1:
                file_path = self.__file_path_manager.add_suffix(file_path, index)
        else:
            file_path = None
        save_show_figure(fig, show, file_path)
        return fig

    def _pdf(
        self,
        index,  # type: int
    ):
        # type: (...) -> Callable
        """Get the probability density function of a marginal.

        Args:
            index: The index of a component of the random variable.

        Return:
            The probability density function
                of the given component of the random variable.
        """

        def pdf(
            point,  # type: float
        ):
            # type: (...) -> float
            """Probability Density Function (PDF).

            Args:
                point: An evaluation point.

            Returns:
                The PDF value at the evaluation point.
            """
            raise NotImplementedError

        return pdf

    def _cdf(
        self,
        index,  # type: int
    ):
        # type: (...) -> Callable
        """Get the cumulative density function of a marginal.

        Args:
            index: The index of a component of the random variable.

        Return:
            The cumulative density function
                of the given component of the random variable.
        """

        def cdf(
            level,  # type: float
        ):
            # type: (...) -> float
            """Cumulative Density Function (CDF).

            Args:
                level: A probability level.

            Returns:
                The CDF value for the probability level.
            """
            raise NotImplementedError

        return cdf



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo, Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Scaling a variable with a linear transformation.

The :class:`.Scaler` class implements the default scaling method
applying to some parameter :math:`z`:

.. math::

    \\bar{z} := \\text{offset} + \\text{coefficient}\\times z

where :math:`\\bar{z}` is the scaled version of :math:`z`.
This scaling method is a linear transformation
parameterized by an offset and a coefficient.

In this default scaling method,
the offset is equal to 0 and the coefficient is equal to 1.
Consequently,
the scaling operation is the identity: :math:`\\bar{z}=z`.
This method has to be overloaded.

.. seealso::

   :mod:`~gemseo.mlearning.transform.scaler.min_max_scaler`
   :mod:`~gemseo.mlearning.transform.scaler.standard_scaler`
"""
from __future__ import division, unicode_literals

import logging

from numpy import diag, eye, ndarray

from gemseo.mlearning.transform.transformer import Transformer, TransformerFitOptionType

LOGGER = logging.getLogger(__name__)


class Scaler(Transformer):
    """Data scaler."""

    def __init__(
        self,
        name="Scaler",  # type: str
        offset=0.0,  # type: float
        coefficient=1.0,  # type: float
    ):  # type: (...) -> None
        """
        Args:
            name: A name for this transformer.
            offset: The offset of the linear transformation.
            coefficient: The coefficient of the linear transformation.
        """
        super(Scaler, self).__init__(name, offset=offset, coefficient=coefficient)

    @property
    def offset(self):  # type: (...) -> float
        """The scaling offset."""
        return self.parameters["offset"]

    @property
    def coefficient(self):  # type: (...) -> float
        """The scaling coefficient."""
        return self.parameters["coefficient"]

    @offset.setter
    def offset(
        self,
        value,  # type: float
    ):  # type: (...) -> None
        self.parameters["offset"] = value

    @coefficient.setter
    def coefficient(
        self,
        value,  # type: float
    ):  # type: (...) -> None
        self.parameters["coefficient"] = value

    def fit(
        self,
        data,  # type: ndarray
        *args  # type: TransformerFitOptionType
    ):  # type: (...) -> None
        LOGGER.warning(
            (
                "The %s.fit() function does nothing; "
                "the instance of %s uses the coefficient and offset "
                "passed at its initialization"
            ),
            self.__class__.__name__,
            self.__class__.__name__,
        )

    def transform(
        self,
        data,  # type: ndarray
    ):  # type: (...) -> ndarray
        return self.offset + self.coefficient * data

    def inverse_transform(
        self,
        data,  # type: ndarray
    ):  # type: (...) -> ndarray
        return (data - self.offset) / self.coefficient

    def compute_jacobian(
        self,
        data,  # type: ndarray
    ):  # type: (...) -> ndarray
        if not isinstance(self.coefficient, ndarray):
            return self.coefficient * eye(data.shape[-1])
        else:
            return diag(self.coefficient)

    def compute_jacobian_inverse(
        self,
        data,  # type: ndarray
    ):  # type: (...) -> ndarray
        if not isinstance(self.coefficient, ndarray):
            return 1 / self.coefficient * eye(data.shape[-1])
        else:
            return diag(1 / self.coefficient)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""Draw a scatter plot from a :class:`.Dataset`.

A :class:`Scatter` plot represents a set of points
:math:`\{x_i,y_i\}_{1\leq i \leq n}` as markers on a classical plot
where the color of points can be heterogeneous.
"""
from __future__ import division, unicode_literals

from typing import List, Mapping

import matplotlib.pyplot as plt
from matplotlib.figure import Figure

from gemseo.post.dataset.dataset_plot import DatasetPlot


class Scatter(DatasetPlot):
    """Plot curve y versus x."""

    def _plot(
        self,
        properties,  # type: Mapping
        x,  # type: str
        y,  # type: str
        x_comp=0,  # type: str
        y_comp=0,  # type: str
    ):  # type: (...) -> List[Figure]
        """
        Args:
            x: The name of the variable on the x-axis.
            y: The name of the variable on the y-axis.
            x_comp: The component of x.
            y_comp: The component of y.
        """
        color = properties.get(self.COLOR) or "blue"
        x_data = self.dataset[x][x][:, x_comp]
        y_data = self.dataset[y][y][:, y_comp]

        fig = plt.figure()
        axes = fig.add_subplot(1, 1, 1)
        axes.scatter(x_data, y_data, color=color)
        if self.dataset.sizes[x] == 1:
            axes.set_xlabel(self.xlabel or x)
        else:
            axes.set_xlabel(self.xlabel or "{}({})".format(x, x_comp))
        if self.dataset.sizes[y] == 1:
            axes.set_ylabel(self.ylabel or y)
        else:
            axes.set_ylabel(self.ylabel or "{}({})".format(y, y_comp))
        return [fig]



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Class to create an uniform distribution from the OpenTURNS library.

This class inherits from :class:`.OTDistribution`.
"""

from __future__ import division, unicode_literals

from typing import Optional

from gemseo.uncertainty.distributions.openturns.distribution import OTDistribution


class OTUniformDistribution(OTDistribution):
    """Create an uniform distribution.

    Example:
        >>> from gemseo.uncertainty.distributions.openturns.uniform import (
        ...     OTUniformDistribution
        >>> )
        >>> distribution = OTUniformDistribution('x', -1, 1)
        >>> print(distribution)
        Uniform(lower=-1, upper=1)
    """

    def __init__(
        self,
        variable,  # type: str
        minimum=0.0,  # type: float
        maximum=1.0,  # type: float
        dimension=1,  # type: int
        transformation=None,  # type: Optional[str]
        lower_bound=None,  # type: Optional[float]
        upper_bound=None,  # type: Optional[float]
        threshold=0.5,  # type: float
    ):  # noqa: D205,D212,D415
        # type: (...) -> None
        """
        Args:
            variable: The name of the uniform random variable.
            minimum: The minimum of the uniform random variable.
            maximum: The maximum of the uniform random variable.
            dimension: The dimension of the uniform random variable.
            transformation: A transformation
                applied to the random variable,
                e.g. 'sin(x)'. If None, no transformation.
            lower_bound: A lower bound to truncate the distribution.
                If None, no lower truncation.
            upper_bound: An upper bound to truncate the distribution.
                If None, no upper truncation.
            threshold: A threshold in [0,1].
        """
        standard_parameters = {self._LOWER: minimum, self._UPPER: maximum}
        super(OTUniformDistribution, self).__init__(
            variable,
            "Uniform",
            (minimum, maximum),
            dimension,
            standard_parameters,
            transformation,
            lower_bound,
            upper_bound,
            threshold,
        )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Python2 and Python3 compatibility layer."""

import operator
import sys
from typing import Callable, Iterable, List, Tuple

from numpy import array, ndarray
from six import PY2, PY3, string_types  # noqa: F401

if PY2:
    string_dtype = "string"
    xrange = xrange  # noqa: F821
    long = long  # noqa: F821

    from inspect import getargspec as _getfullargspec

    from backports.functools_lru_cache import lru_cache  # noqa: F401
    from pathlib2 import Path  # noqa: F401

    from gemseo.third_party.fastjsonschema import (  # noqa: F401
        compile as compile_schema,
    )
    from gemseo.third_party.fastjsonschema.exceptions import (  # noqa: F401
        JsonSchemaException,
    )

    getargspec = _getfullargspec

    def strings_to_unicode_list(
        iterable,  # type: Iterable[str]
    ):  # type: (...) -> List[str]
        """Convert a list of strings to a list of unicode strings."""
        return [s.decode("utf-8") for s in iterable]

    import backports.unittest_mock

    backports.unittest_mock.install()
    from unittest import mock


else:
    string_dtype = "bytes"
    long = int
    xrange = range

    from functools import lru_cache  # noqa: F401
    from inspect import getfullargspec as _getfullargspec
    from pathlib import Path  # noqa: F401
    from unittest import mock  # noqa: F401

    from fastjsonschema import compile as compile_schema  # noqa: F401
    from fastjsonschema.exceptions import JsonSchemaException  # noqa: F401

    def getargspec(
        func,  # type: Callable
    ):  # type: (...) -> Tuple[str]
        """Get arguments specifications."""
        return _getfullargspec(func)[:4]

    def strings_to_unicode_list(
        iterable,  # type: Iterable[str]
    ):  # type: (...) -> Iterable[str]
        """Convert a list of strings to a list of unicode strings."""
        return iterable


if sys.version_info < (3, 6):
    from collections import OrderedDict  # noqa: F401
else:
    OrderedDict = dict


def string_array(
    data,  # type: ndarray
):  # type: (...) -> ndarray
    """Creates a numpy array of strings from data the dtype is adjusted (bytes in py3,
    string in py2)"""
    return array(data, dtype=string_dtype)


if sys.version_info < (3, 8):

    def accumulate(iterable, func=operator.add, initial=None):
        """Accumulate implementation in plain Python.

        Args:
            iterable: An iterable sequence.
            func: An operator to apply on each element of the sequence.
            initial: The inital value of the accumulator.

        Yields:
            The accumulated item.

        Example:
            >>> accumulate([1,2,3,4,5])
            1 3 6 10 15
            >>> accumulate([1,2,3,4,5], initial=100)
            100 101 103 106 110 115
            >>> accumulate([1,2,3,4,5], operator.mul)
            1 2 6 24 120
        """
        it = iter(iterable)
        total = initial
        if initial is None:
            try:
                total = next(it)
            except StopIteration:
                return
        yield total
        for element in it:
            total = func(total, element)
            yield total


else:
    from itertools import accumulate  # noqa: F401



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Remi Lafage
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Monitoring mechanism to track |g| execution (update events)
***********************************************************
"""
from __future__ import division, unicode_literals

from future.utils import with_metaclass

from gemseo.utils.singleton import SingleInstancePerAttributeId


class Monitoring(with_metaclass(SingleInstancePerAttributeId, object)):
    """This class implements the observer pattern.

    It is a singleton, it is called by |g| core classes like MDODicipline whenever an
    event of interest like a status change occurs. Client objects register with
    add_observer and are notified whenever a discipline status change occurs.
    """

    def __init__(self, scenario):
        """Initializes the monitoring.

        :param scenario: the scenario to be monitored.
        :type scenario: MDOScenario
        """
        self._observers = []
        self.workflow = scenario.get_expected_workflow()
        self.workflow.set_observer(self)
        self.workflow.enable()

    def add_observer(self, observer):
        """Registers an observer object interested in observable update events.

        :param observer: object to be notified
        """
        if observer not in self._observers:
            self._observers.append(observer)

    def remove_observer(self, observer):
        """Unsubscribes the given observer.

        :param observer: observer to be removed
        """
        if observer in self._observers:
            self._observers.remove(observer)

    def remove_all_observers(self):
        """Unsubscribes all observers."""
        self._observers = []

    def update(self, atom):
        """Notifies observers that the corresponding observable object is updated.
        Observers have to know what to retrieve from the observable object.

        :param atom: updated object
        """
        for obs in self._observers:
            obs.update(atom)

    def get_statuses(self):
        """Gets the statuses of all disciplines.

        :returns: a dictionary of all statuses, keys are the atom ids
        :rtype: dict
        """
        return self.workflow.get_state_dict()

    def __str__(self):
        """Returns the string representation of the workflow.

        :returns: string representation of the workflow
        :rtype: str
        """
        return str(self.workflow)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#       :author: Remi Lafage
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Creation of a XDSM diagram from a scenario.

The :class:`.XDSMizer` generates a JSON file.

The latter is used by
the `XDSMjs javascript library <https://github.com/OneraHub/XDSMjs>`_
to produce an interactive web XDSM
and by the pyxdsm python library
to produce TIKZ and PDF versions of the XDSM.

For more information, see:
A. B. Lambe and J. R. R. A. Martins, “Extensions to the Design Structure Matrix for
the Description of Multidisciplinary Design, Analysis, and Optimization Processes”,
Structural and Multidisciplinary Optimization, vol. 46, no. 2, p. 273-284, 2012.
"""
from __future__ import division, unicode_literals

import logging
import webbrowser
from json import dumps
from multiprocessing import RLock
from os.path import basename, splitext
from tempfile import mkdtemp
from typing import Any, Dict, List, Mapping, Optional, Union

from gemseo.core.discipline import MDODiscipline
from gemseo.core.doe_scenario import DOEScenario
from gemseo.core.execution_sequence import (
    AtomicExecSequence,
    CompositeExecSequence,
    LoopExecSequence,
    ParallelExecSequence,
    SerialExecSequence,
)
from gemseo.core.mdo_scenario import MDOScenario, MDOScenarioAdapter
from gemseo.core.monitoring import Monitoring
from gemseo.core.scenario import Scenario
from gemseo.mda.mda import MDA
from gemseo.utils.locks import synchronized
from gemseo.utils.py23_compat import Path
from gemseo.utils.show_utils import generate_xdsm_html
from gemseo.utils.xdsm_to_pdf import xdsm_data_to_pdf

LOGGER = logging.getLogger(__name__)

OPT_NAME = OPT_ID = "Opt"
USER_NAME = USER_ID = "_U_"

EdgeType = Dict[str, Union[MDODiscipline, List[str]]]
NodeType = Dict[str, str]
IdsType = Any

XdsmType = Dict[str, Any]


class XDSMizer(object):
    """Build the XDSM diagram of a scenario as a JSON structure."""

    def __init__(
        self,
        scenario,  # type: Scenario
        hashref="root",  # type: str
        level=0,  # type: int
        expected_workflow=None,  # type: Optional[CompositeExecSequence]
    ):  # type: (...) -> None
        """

        Args:
            scenario: The scenario to be represented as an XDSM diagram.
            hashref: The keyword used in the JSON structure
                to reference the dictionary data structure
                whose keys are "nodes", "edges", "workflow" and "optpb".
            level: The depth of the scenario. Root scenario is level 0.
            expected_workflow: The expected workflow,
                describing the sequence of execution of the different disciplines
                (:class:`.MDODiscipline`, :class:`.Scenario`, :class:`.MDA`, etc.)
        """
        self.scenario = scenario
        self.level = level
        self.hashref = hashref
        self.lock = RLock()
        self._monitor = None
        self.outdir = "."
        self.outfilename = "xdsm.json"
        self.to_hashref = {}
        self.to_id = {}  # dictionary to map AtomicExecSequence to XDSM id
        self.initialize(expected_workflow)
        self.print_statuses = False  # Prints the statuses in the console
        self.latex_output = False

    def initialize(
        self,
        workflow=None,  # type: Optional[CompositeExecSequence]
    ):  # type: (...) -> None
        """Initialize the XDSM from a workflow.

        The initialization also creates sub-XDSM diagram accordingly.

        Args:
            workflow: The composite execution sequence.
                If None, use the scenario's one.
        """
        self.sub_xdsmizers = []
        # Find disciplines from workflow structure
        if workflow:
            self.workflow = workflow
        else:
            self.workflow = self.scenario.get_expected_workflow()
        self.atoms = XDSMizer._get_single_level_atoms(self.workflow)

        self.to_hashref = {}
        level = self.level + 1
        num = 1
        for atom in self.atoms:
            if atom.discipline.is_scenario():
                if atom.discipline == self.scenario:
                    self.to_hashref[atom] = "root"
                    self.root_atom = atom
                else:  # sub-scenario
                    name = atom.discipline.name
                    self.to_hashref[atom] = "{}_scn-{}-{}".format(name, level, num)
                    sub_workflow = XDSMizer._find_sub_workflow(self.workflow, atom)
                    self.sub_xdsmizers.append(
                        XDSMizer(
                            atom.discipline, self.to_hashref[atom], level, sub_workflow
                        )
                    )
                    num += 1

    def monitor(
        self,
        outdir=".",  # type: str
        outfilename="xdsm.json",  # type: str
        print_statuses=False,  # type: str
        latex_output=False,  # type: bool
    ):  # type: (...) -> None
        """Monitor the discipline execution by generating XDSM json file on discipline
        status update.

        Args:
            outdir: The name of the directory to store the different files.
            outfilename: The name of the JSON file.
            print_statuses: If True, print the statuses in the console at each update.
            latex_output: If True, save the XDSM to tikz, tex and pdf files.
        """
        self._monitor = Monitoring(self.scenario)
        self._monitor.add_observer(self)
        # have to reinitialize with monitored workflow
        self.initialize(self._monitor.workflow)
        self.outdir = outdir
        self.outfilename = outfilename
        self.print_statuses = print_statuses
        self.latex_output = latex_output

    def update(
        self,
        atom,  # type: AtomicExecSequence
    ):  # type: (...) -> None  # pylint: disable=unused-argument
        """Generate a new XDSM regarding the atom status update.

        Args:
            atom: The discipline which status is monitored.
        """
        self.run(
            output_directory_path=self.outdir,
            outfilename=self.outfilename,
            latex_output=self.latex_output,
        )
        if self.print_statuses:
            LOGGER.info(str(self._monitor))

    def run(
        self,
        output_directory_path=None,  # type: Optional[str]
        latex_output=False,  # type: bool
        outfilename="xdsm.html",  # type: str
        html_output=True,  # type: bool
        json_output=False,  # type: bool
        open_browser=False,  # type: bool
    ):  # type: (...) -> XdsmType
        """Generate a XDSM diagram from the process.

        By default,
        a self contained HTML file is generated,
        that can be viewed in a browser.

        Args:
            output_directory_path: The name of the directory to store the JSON file.
                If None, the current working directory is used.
                If open_browser is True and outdir is None,
                the file is stored in a temporary directory.
            outfilename: The name of the JSON file.
            latex_output: If True, save the XDSM to tikz, tex and pdf files.
            open_browser: If True, open the web browser and display the XDSM.
            html_output: If True, save the XDSM in a self-contained HTML file
            json_output: If True, save the JSON file.

        Returns:
            The XDSM structure expressed as a dictionary
            whose keys are "nodes", "edges", "workflow" and "optpb".
        """
        xdsm = self.xdsmize()
        xdsm_json = dumps(xdsm, indent=2, ensure_ascii=False)
        base = basename(outfilename)
        outfile_basename = splitext(base)[0]

        no_html_loc = False

        if output_directory_path is None:
            output_directory_path = Path.cwd()
            no_html_loc = True
        else:
            output_directory_path = Path(output_directory_path)

        if json_output:
            json_path = output_directory_path / "{}.json".format(outfile_basename)
            with json_path.open("w") as file_stream:
                file_stream.write(xdsm_json)

        if latex_output:
            xdsm_data_to_pdf(xdsm, output_directory_path, outfile_basename)

        if html_output or open_browser:
            if no_html_loc:
                output_directory_path = Path(mkdtemp(suffix="", prefix="tmp", dir=None))
            out_file_path = (output_directory_path / outfile_basename).with_suffix(
                ".html"
            )
            LOGGER.info("Generating HTML XDSM file in : %s", out_file_path)
            generate_xdsm_html(xdsm, out_file_path)
            if open_browser:
                url = "file://{}".format(out_file_path)
                webbrowser.open(url, new=2)  # open in new tab
            return out_file_path

        return xdsm

    def get_all_sub_xdsmizers(self):  # type: (...) -> List[XDSMizer]
        """Retrieve all the sub-xdsmizers corresponding to the sub-scenarios.

        Returns:
            The sub-xdsmizers.
        """
        result = []
        for sub in self.sub_xdsmizers:
            result.append(sub)
            result.extend(sub.get_all_sub_xdsmizers())
        return result

    @synchronized
    def xdsmize(
        self,
        algoname="Optimizer",  # type: str
    ):  # type: (...) -> Dict[str,Any]
        """Build the data structure to be used to generate the JSON file.

        Args:
            algoname: The name under which a scenario appears in an XDSM.

        Returns:
            The XDSM structure expressed as a dictionary
            whose keys are "nodes", "edges", "workflow" and "optpb".
        """
        nodes = self._create_nodes(algoname)
        edges = self._create_edges()
        workflow = self._create_workflow()
        optpb = str(self.scenario.formulation.opt_problem)

        if self.level == 0:
            res = {
                self.hashref: {
                    "nodes": nodes,
                    "edges": edges,
                    "workflow": workflow,
                    "optpb": optpb,
                }
            }
            for sub_xdsmizer in self.get_all_sub_xdsmizers():
                if sub_xdsmizer.scenario.name.endswith("ing"):
                    name = "{}er".format(sub_xdsmizer.scenario.name[:-3])
                elif sub_xdsmizer.scenario.name.endswith("Scenario"):
                    if isinstance(sub_xdsmizer.scenario, DOEScenario):
                        name = "Trade-Off"
                    elif isinstance(sub_xdsmizer.scenario, MDOScenario):
                        name = "Optimizer"
                    else:
                        name = sub_xdsmizer.scenario.name
                else:
                    name = sub_xdsmizer.scenario.name
                res[sub_xdsmizer.hashref] = sub_xdsmizer.xdsmize(name)
            return res
        return {"nodes": nodes, "edges": edges, "workflow": workflow, "optpb": optpb}

    def _create_nodes(
        self,
        algoname,  # type: str
    ):  # type: (...) ->  List[NodeType]# pylint: disable=too-many-branches
        """Create the nodes of the XDSM from the scenarios and the disciplines.

        Args:
            algoname: The name under which a scenario appears in an XDSM.
        """
        nodes = []
        self.to_id = {}

        statuses = self.workflow.get_state_dict()

        # Optimization
        self.to_id[self.root_atom] = OPT_ID
        opt_node = {"id": OPT_ID, "name": algoname, "type": "optimization"}
        if statuses[self.root_atom.uuid]:
            opt_node["status"] = statuses[self.root_atom.uuid]

        nodes.append(opt_node)

        # Disciplines
        for atom_id, atom in enumerate(
            self.atoms
        ):  # pylint: disable=too-many-nested-blocks
            # if a node already created from an atom with same discipline
            # at one level just reference the same node
            for ref_atom in self.to_id:
                if atom.discipline == ref_atom.discipline:
                    self.to_id[atom] = self.to_id[ref_atom]

                    if (
                        atom.status
                        and atom.parent.status is MDODiscipline.STATUS_RUNNING
                    ):

                        node = None
                        for a_node in nodes:
                            if a_node["id"] == self.to_id[atom]:
                                node = a_node
                                break

                        if not node:
                            # TODO: add specific exception?
                            raise "Node " + self.to_id[
                                atom
                            ] + " not found in " + nodes  # pragma: no cover

                        node["status"] = atom.status

                    break

            if atom in self.to_id:
                continue

            self.to_id[atom] = "Dis" + str(atom_id)
            node = {"id": self.to_id[atom], "name": atom.discipline.name}

            # node type
            if isinstance(atom.discipline, MDA):
                node["type"] = "mda"
            elif atom.discipline.is_scenario():
                node["type"] = "mdo"
                node["subxdsm"] = self.to_hashref[atom]
                node["name"] = self.to_hashref[atom]
            else:
                node["type"] = "analysis"

            if statuses[atom.uuid]:
                node["status"] = statuses[atom.uuid]

            nodes.append(node)

        return nodes

    def _create_edges(self):  # type: (...) -> List[EdgeType]
        """Create the edges of the XDSM from the dataflow of the scenario."""
        edges = []
        # convenient method to factorize code for creating and appending edges

        def add_edge(
            from_edge,  # type: MDODiscipline
            to_edge,  # type: MDODiscipline
            varnames,  # type: List[str]
        ):  # type: (...) -> None
            """Add an edge from a discipline to another with variables names as label.

            Args:
                from_edge: The starting discipline.
                to_edge: The end discipline.
                varnames: The names of the variables
                    going from the starting discipline to the end one.
            """
            edge = {"from": from_edge, "to": to_edge, "name": ", ".join(varnames)}
            edges.append(edge)

        # For User to/from optimization
        opt_pb = self.scenario.formulation.opt_problem

        # fct names such as -y4
        functions_names = opt_pb.get_all_functions_names()

        # output variables used by the fonction (eg y4)
        fct_varnames = [f.outvars for f in opt_pb.get_all_functions()]
        function_varnames = []
        for fvars in fct_varnames:
            function_varnames.extend(fvars)

        to_user = functions_names
        to_opt = self.scenario.get_optim_variables_names()

        user_pattern = "L({})" if self.scenario.name == "Sampling" else "{}^(0)"
        opt_pattern = "{}^(1:N)" if self.scenario.name == "Sampling" else "{}^*"
        add_edge(USER_ID, OPT_ID, [user_pattern.format(x) for x in to_opt])
        add_edge(OPT_ID, USER_ID, [opt_pattern.format(x) for x in to_user])

        # Disciplines to/from optimization
        for atom in self.atoms:
            if atom is not self.root_atom:
                varnames = set(atom.discipline.get_input_data_names()) & set(
                    self.scenario.get_optim_variables_names()
                )
                if varnames:
                    add_edge(OPT_ID, self.to_id[atom], varnames)

                varnames = set(atom.discipline.get_output_data_names()) & set(
                    function_varnames
                )
                # print set(disc.get_output_data_names()), set(functions_names)
                if varnames:
                    add_edge(self.to_id[atom], OPT_ID, varnames)

        # Disciplines to User/Optimization (from User is already handled at
        # optimizer level)
        disc_to_opt = function_varnames
        for atom in self.atoms:
            if atom is not self.root_atom:
                # special case MDA : skipped
                if isinstance(atom.discipline, MDA):
                    continue
                out_to_user = [
                    o
                    for o in atom.discipline.get_output_data_names()
                    if o not in disc_to_opt
                ]
                out_to_opt = [
                    o
                    for o in atom.discipline.get_output_data_names()
                    if o in disc_to_opt
                ]
                if out_to_user:
                    add_edge(self.to_id[atom], USER_ID, [x + "^*" for x in out_to_user])
                if out_to_opt:
                    add_edge(self.to_id[atom], OPT_ID, out_to_opt)

        # Disciplines to/from disciplines
        for coupling in self.scenario.get_expected_dataflow():
            (disc1, disc2, varnames) = coupling
            add_edge(
                self.to_id[self._find_atom(disc1)],
                self.to_id[self._find_atom(disc2)],
                varnames,
            )

        return edges

    @staticmethod
    def _get_single_level_atoms(
        workflow,  # type: CompositeExecSequence
    ):  # type: (...) -> List[AtomicExecSequence]
        """Retrieve the list of atoms of the given workflow.

        This method does not look into the loop execution sequences
        coming from the scenario.
        Thus, it retrieves the atoms for a one level XDSM diagram.

        Args:
            The composite execution sequence.

        Returns:
            The atomic execution sequences.
        """
        atoms = []
        for seq in workflow.sequence_list:
            if isinstance(seq, LoopExecSequence):
                atoms.append(seq.atom_controller)
                if not seq.atom_controller.discipline.is_scenario():
                    atoms += XDSMizer._get_single_level_atoms(seq.iteration_sequence)
            elif isinstance(seq, AtomicExecSequence):
                atoms.append(seq)
            else:
                atoms += XDSMizer._get_single_level_atoms(seq)
        return atoms

    def _find_atom(
        self,
        discipline,  # type: MDODiscipline
    ):  # type: (...) -> AtomicExecSequence
        """Find the atomic sequence corresponding to a given discipline.

        Args:
            discipline: A discipline.

        Returns:
            The atomic sequence corresponding to the given discipline.

        Raises:
            ValueError: If the atomic sequence is not found.
        """
        atom = None
        if isinstance(discipline, MDOScenarioAdapter):
            atom = self._find_atom(discipline.scenario)
        else:
            for atom_i in self.atoms:
                if discipline == atom_i.discipline:
                    atom = atom_i
        if atom is None:
            disciplines = [a.discipline for a in self.atoms]
            raise ValueError(
                "Discipline {} not found in {}".format(discipline, disciplines)
            )
        return atom

    @staticmethod
    def _find_sub_workflow(
        workflow,  # type: CompositeExecSequence
        atom_controller,  # type: AtomicExecSequence
    ):  # type: (...) -> Optional[LoopExecSequence]
        """Find the sub-workflow from a workflow and controller atom in it.

        Args:
            workflow: The workflow from which to find a sub-workflow.
            atom_controller: The atomic execution sequence that controls
                the loop execution sequence to find.

        Returns:
            The sub-workflow.
            None if the list of execution sequences of the original workflow is empty.
        """
        sub_workflow = None
        for seq in workflow.sequence_list:
            if isinstance(seq, LoopExecSequence):
                if seq.atom_controller.uuid == atom_controller.uuid:
                    sub_workflow = seq
                    return sub_workflow

                sub_workflow = sub_workflow or XDSMizer._find_sub_workflow(
                    seq.iteration_sequence, atom_controller
                )
            elif not isinstance(seq, AtomicExecSequence):
                sub_workflow = sub_workflow or XDSMizer._find_sub_workflow(
                    seq, atom_controller
                )

        return sub_workflow

    def _create_workflow(self):  # type: (...) -> List[str,IdsType]
        """Manage the creation of the XDSM workflow creation from a formulation one."""
        workflow = [USER_ID, expand(self.workflow, self.to_id)]
        return workflow


def expand(
    wks,  # type: CompositeExecSequence
    to_id,  # type: Mapping[str,str]
):  # type: (...) -> IdsType
    """Expand the workflow structure as an ids structure using to_id mapping.

    The expansion preserve the structure
    while replacing the object by its id in all case
    except when a tuple is encountered as cdr
    then the expansion transforms loop[A, (B,C)] in [idA, {'parallel': [idB, idC]}].

    Args:
        wks: The workflow structure.
        to_id: The mapping dict from object to id.

    Returns:
        The ids structure valid to be used as XDSM json chains.
    """
    if isinstance(wks, SerialExecSequence):
        res = []
        for seq in wks.sequence_list:
            res += expand(seq, to_id)
        ids = res
    elif isinstance(wks, ParallelExecSequence):
        res = []
        for seq in wks.sequence_list:
            res += expand(seq, to_id)
        ids = [{"parallel": res}]
    elif isinstance(wks, LoopExecSequence):
        if (
            wks.atom_controller.discipline.is_scenario()
            and to_id[wks.atom_controller] != OPT_ID
        ):
            # sub-scnario consider only the controller
            ids = [to_id[wks.atom_controller]]
        else:
            ids = [to_id[wks.atom_controller], expand(wks.iteration_sequence, to_id)]
    elif isinstance(wks, AtomicExecSequence):
        ids = [to_id[wks]]
    else:
        raise Exception("Bad execution sequence: found {}".format(wks))
    return ids



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation
#               and/or initial documentation
#        :author: Sobieski, Agte, and Sandusky
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
#        :author: Damien Guenot
#        :author: Francois Gallard
# From NASA/TM-1998-208715
# Bi-Level Integrated System Synthesis (BLISS)
# Sobieski, Agte, and Sandusky
"""
SSBJ base class
***************
"""
from __future__ import division, unicode_literals

import cmath
import logging
import math

from numpy import array, asarray, complex128, concatenate, dot, float64, zeros
from numpy.linalg import lstsq
from scipy import linalg
from six import string_types

LOGGER = logging.getLogger(__name__)


DEG_TO_RAD = math.pi / 180.0


class SobieskiBase(object):
    """Class defining Sobieski problem and related method to the problem such as
    disciplines computation, constraints, reference optimum."""

    DTYPE_COMPLEX = "complex128"
    DTYPE_DOUBLE = "float64"
    DTYPE_DEFAULT = DTYPE_COMPLEX

    def __init__(self, dtype):
        """Constructor.

        :param dtype: data type
        :type dtype: str
        """
        self.dtype = dtype
        if dtype == complex128:
            self.math = cmath
        elif dtype == float64:
            self.math = math
        elif dtype == self.DTYPE_DOUBLE:
            self.math = math
            self.dtype = float64
        elif dtype == self.DTYPE_COMPLEX:
            self.math = cmath
            self.dtype = complex128
        else:
            raise ValueError("Unknown dtype : " + str(dtype))

        self.i_0 = None

    def get_default_x0(self):
        """Return a default initial value for design variables.

        :returns: initial design variables
        :rtype: numpy array

        warning:
        ##DO NOT CHANGE VALUE: THEY ARE USED FOR POLYNOMIAL APPROXIMATION ##
        """
        return array(
            [0.25, 1.0, 1.0, 0.5, 0.05, 45000.0, 1.6, 5.5, 55.0, 1000.0],
            dtype=self.dtype,
        )

    def default_constants(self):
        """Definition of constants vector constants for Sobieski problem.

        :returns: constant vector
        :rtype: numpy array
        """
        constants = zeros(5, dtype=self.dtype)

        # Constants of problem
        constants[0] = 2000.0  # minimum fuel weight
        constants[1] = 25000.0  # miscellaneous weight
        constants[2] = 6.0  # Maximum load factor
        constants[3] = 4360.0  # Engine weight reference
        constants[4] = 0.01375  # Minimum drag coefficient
        return constants

    @staticmethod
    def get_sobieski_bounds_tuple():
        """Set the input design bounds and return them as a tuple of tuples.

        :returns: Subtuple is build with
            lower bound and upper bound.
        :rtype: tuple(tuple)
        """
        bounds_tuple = (
            (0.1, 0.4),
            (0.75, 1.25),
            (0.75, 1.25),
            (0.1, 1),
            (0.01, 0.09),
            (30000.0, 60000.0),
            # (0.01, 0.09) with threshold, (0.038, 0.06)
            # without threshold
            (1.4, 1.8),
            (2.5, 8.5),
            (40.0, 70.0),
            (500.0, 1500.0),
        )

        return bounds_tuple

    @classmethod
    def get_sobieski_bounds(cls):
        """Set the input design bounds and return them as 2 numpy arrays.

        :returns: upper and lower bounds
        :rtype: numpy array, numpy array
        """
        bounds_tuple = cls.get_sobieski_bounds_tuple()
        bounds_array = asarray(bounds_tuple)
        upper_bounds = bounds_array[:, 1]
        lower_bounds = bounds_array[:, 0]
        return upper_bounds, lower_bounds

    @classmethod
    def get_bounds_by_name(cls, variables_names):
        """Return bounds of design variables and coupling variables.

        :param variables_names: names of variables
        :type variables_names: str or list(str)
        :returns: lower bound and upper bound
        """

        if isinstance(variables_names, string_types):
            variables_names = [variables_names]
        bounds_tuple = cls.get_sobieski_bounds_tuple()
        bounds_dict = {
            "x_1": array(bounds_tuple[0:2]),
            "x_2": array([bounds_tuple[2]]),
            "x_3": array([bounds_tuple[3]]),
            "x_shared": array(bounds_tuple[4:10]),
            "y_14": array([(4.97e04, 5.14e04), (-1.54e04, 3.0e04)]),
            "y_12": array([(4.97e04, 5.15e04), (0.9, 1.00)]),
            "y_21": array([(4.97e04, 5.15e04)]),
            "y_23": array([(6.73e03, 1.76e04)]),
            "y_24": array([(0.88, 7.42)]),
            "y_31": array([(5.92e03, 6.79e03)]),
            "y_32": array([(0.47, 0.53)]),
            "y_34": array([(0.88, 1.32)]),
        }

        for yuk in ["y_21", "y_14", "y_12", "y_23", "y_24", "y_31", "y_32", "y_34"]:
            bounds_dict[yuk][:, 0] = bounds_dict[yuk][:, 0] * 0.5
            bounds_dict[yuk][:, 1] = bounds_dict[yuk][:, 1] * 1.5

        bounds = bounds_dict[variables_names[0]]
        #         LOGGER.debug(15 * " " + "get_bounds_by_name")
        for var_name in variables_names[1:]:
            #             LOGGER.debug("var_name " + var_name)
            bounds = concatenate((bounds, bounds_dict[var_name]), axis=0)
        return bounds[:, 0], bounds[:, 1]

    def __compute_mtx_shifted(self, s_bound, index):
        """Compute a matrix of shifted values of design variables.

        :param s_bound: vector of bounds used to control slope of
                polynomial function
        :type s_bound: numpy array
        :param index: index of design variable in polynomial function
        :type index: integer
        :returns: mtx_shifted
        :rtype: numpy array
        """
        #         s_mid = 0.0  # independent variable mid point
        #         s_lower = s_mid - s_bound[index]
        #         s_upper = s_mid + s_bound[index]
        s_lower = -s_bound[index]
        s_upper = s_bound[index]
        s_lower_2 = s_lower * s_lower
        #         So_2 = s_mid * s_mid
        s_upper_2 = s_upper * s_upper
        mtx_shifted = array(
            [
                [1.0, s_lower, s_lower_2],
                #             [1.0, s_mid, s_mid_2],
                [1.0, 0.0, 0.0],
                [1.0, s_upper, s_upper_2],
            ],
            dtype=self.dtype,
        )
        return mtx_shifted

    @staticmethod
    def __compute_a(mtx_shifted, f_bound, ao_coeff, ai_coeff, aij_coeff, index):
        """Compute the interpolation terms.

        :returns: ao_coeff, ai_coeff, aij_coeff
            constant, linear and quadratic terms
        :rtype: float, 1D array, 2D array (numpy)
        """
        try:
            a_mat = linalg.solve(mtx_shifted, f_bound)
        except linalg.LinAlgError:
            LOGGER.warning(
                "Sobieski polynomial approximation: "
                "exact linear system solve failed, using "
                " approximate least squares method instead."
            )
            a_mat, _, _, _ = lstsq(mtx_shifted, f_bound)

        ao_coeff = a_mat[0]
        ai_coeff[index] = a_mat[1][0]
        aij_coeff[index, index] = a_mat[2][0]
        return ao_coeff, ai_coeff, aij_coeff

    def __update_aij(self, aij_coeff, imax):
        """Update of quadratic interpolation terms.

        :param aij_coeff: array of quadratic terms
        :type aij_coeff: numpy array
        :param imax: size of interpolation matrix
        :type imax: integer
        :returns: aij_coeff (modified quadratic terms)
        :rtype: numpy array
        """
        coeff_mtrix = array(
            [
                [0.2736, 0.3970, 0.8152, 0.9230, 0.1108],
                [0.4252, 0.4415, 0.6357, 0.7435, 0.1138],
                [0.0329, 0.8856, 0.8390, 0.3657, 0.0019],
                [0.0878, 0.7248, 0.1978, 0.0200, 0.0169],
                [0.8955, 0.4568, 0.8075, 0.9239, 0.2525],
            ],
            dtype=self.dtype,
        )

        for i in range(imax):
            for j in range(i + 1, imax):
                aij_coeff[i, j] = aij_coeff[i, i] * coeff_mtrix[i, j]
                aij_coeff[j, i] = aij_coeff[i, j]
        return aij_coeff

    def __compute_fbound(self, flag, s_shifted, a_coeff, b_coeff, index):
        """Compute right-hand side of polynomial function system.

        :param flag: functional relationship between var
        :type flag: numpy array
        :param s_shifted: vector of normalized values of independent variables
                shifted around origin
        :type s_shifted: numpy array
        """
        if flag[index] == 5:
            f_bound = array(
                [
                    [1.0 + (0.25 * a_coeff * a_coeff)],
                    [1.0],
                    [1 + 0.25 * b_coeff * b_coeff],
                ],
                dtype=self.dtype,
            )
        else:
            if flag[index] == 0:
                s_shifted = 0.0
            elif flag[index] == 3:
                a_coeff = -a_coeff
                b_coeff = a_coeff
            elif flag[index] == 2:
                b_coeff = 2 * a_coeff
            elif flag[index] == 4:
                a_coeff = -a_coeff
                b_coeff = 2 * a_coeff
            f_bound = array(
                [[1.0 - (0.5 * a_coeff)], [1.0], [1 + 0.5 * b_coeff]], dtype=self.dtype
            )
        return s_shifted, f_bound

    @staticmethod
    def _normalize_s(s_ref, s_new):
        """Normalization of input variables for use of polynomial approximation.

        :param s_ref: vector of initial values of independent
            variables (5 variables at max)
        :type s_ref: numpy array
        :param s_new: vector of current values of
            independent variables
        :type s_new: numpy array
        :returns: normalized value and normalized+centered value
        :rtype: numpy array
        """
        s_norm = s_new / s_ref
        s_norm[s_norm > 1.25] = 1.25
        s_norm[s_norm < 0.75] = 0.75
        s_shifted = s_norm - 1.0  # Shift S near origin
        return s_norm, s_shifted

    def derive_normalize_s(self, s_ref, s_new):
        """Derivation of normalization of input variables for use of polynomial
        approximation.

        :param s_ref: vector of initial values of independent
            variables (5 variables at max)
        :type s_ref: numpy array
        :param s_new: vector of current values of
            independent variables
        :type s_new: numpy array
        :returns: derivatives of normalized value and normalized+centered value
        :rtype: numpy array
        """
        if self.dtype == complex128:
            s_norm = s_new.real / s_ref.real
        else:
            s_norm = s_new / s_ref

        if s_norm > 1.25:
            derive_s_norm = 0.0
        elif s_norm < 0.75:
            derive_s_norm = 0.0
        else:
            derive_s_norm = 1.0 / s_ref
        #        derive_s_norm = s_new / s_ref
        #        derive_s_norm[derive_s_norm > 1.25] = 0.0
        #        derive_s_norm[derive_s_norm < 0.75] = 0.0
        #        derive_s_norm = derive_s_norm / s_ref

        return derive_s_norm

    def derive_poly_approx(self, s_ref, s_new, flag, s_bound):
        """Compute the polynomial coefficients to characterize the behavior of certain
        synthetic variables and function modifiers. Compared to poly_approx, also
        returns polynomial coeff for linearization.

        :param s_ref: vector of initial values of independent
            variables (5 variables at max)
        :type s_ref: numpy array
        :param s_new: vector of current values of
            independent variables
        :type s_new: numpy array
        :param flag: indicates functional relationship between variables:

            - flag = 1: linear >0
            - flag = 2: nonlinear >0
            - flag = 3: linear < 0
            - flag = 4: nonlinear <0
            - flag = 5: parabolic

        :type flag: int
        :param s_bound: offset value for normalization
        :type s_bound: numpy array
        :returns: poly_value: value of synthetic variable or modifier
        :rtype: numpy array
        """

        imax = s_ref.shape[0]
        # Normalize new S with initial
        _, s_shifted = self._normalize_s(s_ref, s_new)
        #       derive_S_norm = self._derive_normalize_S(s_ref, s_new)

        ao_coeff = 0.0
        ai_coeff = zeros(imax, dtype=self.dtype)
        aij_coeff = zeros((imax, imax), dtype=self.dtype)
        for i in range(imax):
            a_coeff = 0.1
            b_coeff = a_coeff
            mtx_shifted = self.__compute_mtx_shifted(s_bound, i)

            s_shifted, f_bound = self.__compute_fbound(
                flag, s_shifted, a_coeff, b_coeff, i
            )

            ao_coeff, ai_coeff, aij_coeff = self.__compute_a(
                mtx_shifted, f_bound, ao_coeff, ai_coeff, aij_coeff, i
            )

        aij_coeff = self.__update_aij(aij_coeff, imax)

        poly_value = (
            ao_coeff
            + dot(ai_coeff, s_shifted)
            + 0.5 * dot(dot(s_shifted, aij_coeff[:imax, :imax]), s_shifted)
        )
        return poly_value[0], ai_coeff, aij_coeff[:imax, :imax], s_shifted

    def poly_approx(self, s_ref, s_new, flag, s_bound):
        """This function calculates polynomial coefficients to characterize the behavior
        of certain synthetic variables and function modifiers.

        :param s_ref: vector of initial values of independent
            variables (5 variables at max)
        :type s_ref: numpy array
        :param s_new: vector of current values of
            independent variables
        :type s_new: numpy array
        :param flag: indicates functional relationship between variables:

            - flag = 1: linear >0
            - flag = 2: nonlinear >0
            - flag = 3: linear < 0
            - flag = 4: nonlinear <0
            - flag = 5: parabolic

        :type flag: int
        :param s_bound: offset value for normalization
        :type s_bound: numpy array
        :returns: poly_value: value of synthetic variable or modifier
        :rtype: numpy array
        """

        imax = s_ref.shape[0]
        # Normalize new S with initial
        _, s_shifted = self._normalize_s(s_ref, s_new)
        # - ai_coeff: vector of coeff for 2nd term
        #         - aij_coeff: matrix of coeff for 3rd term
        #         - ao_coeff: scalar coeff
        ao_coeff = 0.0
        ai_coeff = zeros(imax, dtype=self.dtype)
        aij_coeff = zeros((imax, imax), dtype=self.dtype)
        for i in range(imax):
            a_coeff = 0.1
            b_coeff = a_coeff
            mtx_shifted = self.__compute_mtx_shifted(s_bound, i)

            s_shifted, f_bound = self.__compute_fbound(
                flag, s_shifted, a_coeff, b_coeff, i
            )

            ao_coeff, ai_coeff, aij_coeff = self.__compute_a(
                mtx_shifted, f_bound, ao_coeff, ai_coeff, aij_coeff, i
            )

        aij_coeff = self.__update_aij(aij_coeff, imax)

        poly_value = (
            ao_coeff
            + dot(ai_coeff, s_shifted)
            + 0.5 * dot(dot(s_shifted, aij_coeff[:imax, :imax]), s_shifted)
        )
        return poly_value[0]

    def compute_half_span(self, x_shared):
        """Compute half-span from surface and aspect ratio.

        :param x_shared: global design variables
        :type x_shared: numpy array
        :returns: half-span
        :rtype: numpy array
        """
        return self.math.sqrt(x_shared[3] * x_shared[5]) * 0.5  # 1/2 span

    def compute_thickness(self, x_shared):
        """Compute a wing thickness.

        :param x_shared: global design variables
        :type x_shared: numpy array
        :returns: thickness
        :rtype: numpy array
        """
        return (
            x_shared[0] * x_shared[5] / (self.math.sqrt(x_shared[5] * x_shared[3]))
        )  # thickness

    @staticmethod
    def compute_aero_center(x_1):
        """Computes the aerodynamic center.

        :param x_1: local design variables (structure)
        :type x_1: numpy array
        :returns: aerodynamic center
        :rtype: numpy array
        """
        return (1.0 + 2.0 * x_1[0]) / (3.0 * (1 + x_1[0]))

    def get_initial_values(self):
        """Get initial values used by polynomial functions.

        :returns: initial values
        :rtype: tuple(ndarray)
        """
        self.i_0 = self.get_default_x0()
        x_initial = self.i_0[1]
        tc_initial = self.i_0[4]
        half_span_initial = self.math.sqrt(self.i_0[7] * self.i_0[9]) * 0.5
        aero_center_initial = (1 + 2 * self.i_0[0]) / (3 * (1 + self.i_0[0]))
        cf_initial = self.i_0[2]
        mach_initial = self.i_0[6]
        h_initial = self.i_0[5]
        throttle_initial = self.i_0[3]
        lift_initial = self.dtype(1.0)
        twist_initial = self.dtype(1.0)
        esf_initial = self.dtype(1.0)
        return (
            x_initial,
            tc_initial,
            half_span_initial,
            aero_center_initial,
            cf_initial,
            mach_initial,
            h_initial,
            throttle_initial,
            lift_initial,
            twist_initial,
            esf_initial,
        )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Damien Guenot
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
#         Francois Gallard : refactoring for v1, May 2016
"""
SNOPT optimization library wrapper
**********************************
"""

from __future__ import division, unicode_literals

import logging

from numpy import append, array, concatenate
from numpy import float as np_float
from numpy import float64, hstack
from numpy import int as np_int
from numpy import isinf, ones, reshape
from numpy import str as np_str
from numpy import vstack, where, zeros
from optimize.snopt7 import SNOPT_solver

from gemseo.algos.opt.opt_lib import OptimizationLibrary

LOGGER = logging.getLogger(__name__)
INFINITY = 1e30


class SnOpt(OptimizationLibrary):
    """SNOPT optimization library interface.

    See OptimizationLibrary.
    """

    LIB_COMPUTE_GRAD = False
    OPTIONS_MAP = {"max_iter": "Iteration_limit"}

    MESSAGES_DICT = {
        1: "optimality conditions satisfied",
        2: "feasible point found",
        3: "requested accuracy could not be achieved",
        11: "infeasible linear constraints",
        12: "infeasible linear equalities",
        13: "nonlinear infeasibilities minimized",
        14: "infeasibilities minimized",
        21: "unbounded objective",
        22: "constraint violation limit reached",
        31: "iteration limit reached",
        32: "major iteration limit reached",
        33: "the superbasics limit is too small",
        41: "current point cannot be improved ",
        42: "singular basis",
        43: "cannot satisfy the general constraints",
        44: "ill-conditioned null-space basis",
        51: "incorrect objective derivatives",
        52: "incorrect constraint derivatives",
        61: "undefined function at the first feasible point",
        62: "undefined function at the initial point",
        63: "unable to proceed into undefined region",
        72: "terminated during constraint evaluation",
        73: "terminated during objective evaluation",
        74: "terminated from monitor routine",
        81: "work arrays must have at least 500 elements",
        82: "not enough character storage",
        83: "not enough integer storage",
        84: "not enough real storage",
        91: "invalid input argument",
        92: "basis file dimensions do not match this problem",
        141: "wrong number of basic variables",
        142: "error in basis package",
    }

    def __init__(self):
        """Constructor.

        Generate the library dict, contains the list
        of algorithms with their characteristics:
         * does it require gradient
         * does it handle equality constraints
         * does it handle inequality constraints
        """
        super(SnOpt, self).__init__()
        self.__n_ineq_constraints = 0
        self.__n_eq_constraints = 0
        self.lib_dict = {
            "SNOPTB": {
                self.INTERNAL_NAME: "SNOPTB",
                self.REQUIRE_GRAD: True,
                self.HANDLE_EQ_CONS: True,
                self.HANDLE_INEQ_CONS: True,
                self.DESCRIPTION: "Sparse Nonlinear OPTimizer (SNOPT)",
                self.WEBSITE: "https://ccom.ucsd.edu/~optimizers",
            }
        }

    def _get_options(
        self,
        ftol_rel=1e-9,
        ftol_abs=1e-9,
        xtol_rel=1e-9,
        xtol_abs=1e-9,
        max_time=0,
        max_iter=999,  # pylint: disable=W0221
        normalize_design_space=True,
        **kwargs
    ):
        """Sets the options.

        :param ftol_abs: Objective function tolerance
        :type ftol_abs: float
        :param xtol_abs: Design parameter tolerance
        :type xtol_abs: float
        :param iprint: Default value = 1000)
        :type iprint: int
        :param max_time: Maximum time
        :type max_time: float
        :param max_iter: Default value = 999)
        :type max_iter: int
        :param kwargs: additional options
        :type kwargs: kwargs
        """
        nds = normalize_design_space
        return self._process_options(
            max_iter=max_iter,
            normalize_design_space=nds,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            xtol_abs=xtol_abs,
            max_time=max_time,
            **kwargs
        )

    @staticmethod
    def __eval_func(func, xn_vect):
        """Evaluates a function Trys to call it, if it fails, returns a -1 status.

        :param func: the function to call
        :param xn_vect : the arguments for evaluation
        :returns: function value
        :rtype: numpy
        """
        try:
            val = func(xn_vect)
            return val, 1
        # SNOPT can handle non-computable points
        except Exception as func_error:
            LOGGER.error("SNOPT failed to evaluate the function !")
            LOGGER.error(func_error.args[0])
        return array([float("nan")]), -1

    # cb_ means callback and avoids pylint to raise warnings
    # about unused arguments
    def cb_opt_objective_snoptb(self, mode, nn_obj, xn_vect, n_state=0):
        r"""Objective function + objective gradient  of the optimizer for snOpt
        (from web.stanford.edu/group/SOL/guides/sndoc7.pdf)

        :param mode: indicates whether obj or gradient or both must
            be assigned during the present call of function
            (0 :math:`\leq` mode :math:`\leq` 2).
            mode = 2, assign obj and the known components of gradient
            mode = 1, assign the known components of gradient. obj is ignored.
            mode = 0, only obj need be assigned; gradient is ignored.
        :param nn_obj: number of design variables
        :param xn_vect: normalized design vector
        :param n_state: indicator for first and last call to current function
            n_state = 0: NTR
            n_state = 1: first call to driver.cb_opt_objective_snoptb
            n_state > 1, snOptB is calling subroutine for the last time and:
            n_state = 2       and the current x is optimal
            n_state  = 3, the problem appears to be infeasible
            n_state  = 4, the problem appears to be unbounded;
            n_state  = 5,  an iterations limit was reached. (Default value = 0)
        :returns: status: may be used to indicate that you
            are unable to evaluate
            obj_f or its gradients at the current x.
            (For example, the problem functions may not be defined there).
            obj_f, objective function (except perhaps if mode = 1)
            objective jacobian array (except perhaps if mode = 0)
        :rtype: integer, np array, np array

        """
        obj_func = self.problem.objective
        status = -1
        if mode == 0:
            obj_f, status = self.__eval_func(obj_func.func, xn_vect)
            obj_df = ones((xn_vect.shape[0],)) * 666.0

        if mode == 1:
            obj_df, status = self.__eval_func(obj_func.jac, xn_vect)
            obj_f = -666.0

        if mode == 2:
            obj_df, f_status = self.__eval_func(obj_func.jac, xn_vect)
            obj_f, df_status = self.__eval_func(obj_func.func, xn_vect)
            if f_status == 1 and df_status == 1:
                status = 1
            else:
                status = -1
        return status, obj_f, obj_df

    def __snoptb_create_c(self, xn_vect):
        """Private function that returns evaluation of constraints at design vector
        xn_vect.

        :param xn_vect: normalized design variable vector
        :returns: evaluation of constraints at xn_vect
        :rtype: numpy array
        """
        cstr = array([])
        for constraint in self.problem.get_eq_constraints():
            c_val, status = self.__eval_func(constraint, xn_vect)
            if status == -1:
                return c_val, -1
            cstr = hstack((cstr, c_val))

        for constraint in self.problem.get_ineq_constraints():
            c_val, status = self.__eval_func(constraint, xn_vect)
            if status == -1:
                return c_val, -1
            cstr = hstack((cstr, c_val))
        return cstr, 1

    def __snoptb_create_dc(self, xn_vect):
        """Private function that returns evaluation of constraints gradient at design
        vector xn_vect.

        :param xn_vect: normalized design variable vector
        :returns: evaluation of constraints gradient at xn_vect
        :rtype: numpy array
        """
        dcstr = array([])
        # First equality constraints then inequality
        for constraint in self.problem.get_eq_constraints():
            dc_val, status = self.__eval_func(constraint.jac, xn_vect)
            if status == -1:
                return dc_val, status
            if dcstr:
                dcstr = vstack((dcstr, dc_val))
            else:
                dcstr = dc_val
        for constraint in self.problem.get_ineq_constraints():
            dc_val, status = self.__eval_func(constraint.jac, xn_vect)
            if status == -1:
                return dc_val, status
            if dcstr.size > 0:
                dcstr = vstack((dcstr, dc_val))
            else:
                dcstr = dc_val

        if len(dcstr.shape) > 1:
            dcstr = reshape(dcstr.T, dcstr.shape[0] * dcstr.shape[1])

        return dcstr, 1

    # cb_ means callback and avoids pylint to raise warnings
    # about unused arguments
    def cb_opt_constraints_snoptb(self, mode, nn_con, nn_jac, ne_jac, xn_vect, n_state):
        """Constraints function + constraints gradient of the optimizer for snOpt (from
        web.stanford.edu/group/SOL/guides/sndoc7.pdf)

        :param mode: indicates whether obj or gradient or both must
            be assigned during the present call of function (0 ≤ mode ≤ 2).
            mode = 2, assign obj and the known components of gradient
            mode = 1, assign the known components of gradient. obj is ignored.
            mode = 0, only obj need be assigned; gradient is ignored.
        :param nn_con: number of non-linear constraints
        :param nn_jac: number of dv involved in non-linear
            constraint functions
        :param ne_jac: number of non-zero elements in constraints gradient.
            dcstr is 2D
            ==> ne_jac = nn_con*nn_jac
        :param xn_vect: normalized design vector
        :param n_state: indicator for first and last call to current function
            n_state = 0: NTR
            n_state = 1: first call to driver.cb_opt_objective_snoptb
            n_state > 1, snOptB is calling subroutine for the last time and:
            n_state = 2       and the current x is optimal
            n_state  = 3, the problem appears to be infeasible
            n_state  = 4, the problem appears to be unbounded;
            n_state  = 5,  an iterations limit was reached.
        :returns: status: may be used to indicate that you are unable to
            evaluate
            cstr or its gradients at the current x.
            (For example, the problem functions may not be defined there).
            cstr: constraints function (except perhaps if mode = 1)
            dcstr constraints jacobian array (except perhaps if mode = 0)
        :rtype: integer, np array, np array
        """
        if mode == 0:
            cstr, status = self.__snoptb_create_c(xn_vect)
            dcstr = (
                ones(
                    (
                        (self.__n_eq_constraints + self.__n_ineq_constraints)
                        * xn_vect.shape[0]
                    )
                )
                * 666.0
            )
            status = 1

        elif mode == 1:
            dcstr, status = self.__snoptb_create_dc(xn_vect)
            cstr = ones((self.__n_eq_constraints + self.__n_ineq_constraints,)) * 666.0

        elif mode == 2:
            cstr, c_status = self.__snoptb_create_c(xn_vect)
            dcstr, dc_status = self.__snoptb_create_dc(xn_vect)
            if c_status == 1 and dc_status == 1:
                status = 1
            else:
                status = -1
        return status, cstr, dcstr

    # cb_ means callback and avoids pylint to raise warnings
    # about unused arguments
    @staticmethod
    def cb_snopt_dummy_func(mode, nn_con, nn_jac, ne_jac, xn_vect, n_state):
        """Dummy function required for unconstrained problem.

        :param mode: param nn_con:
        :param nn_jac: param ne_jac:
        :param xn_vect: param n_state:
        :param nn_con:
        :param ne_jac:
        :param n_state:
        """
        return 1.0

    def __preprocess_snopt_constraints(self, names):
        """Private function to set snopt parameters according to constraints.

        :param names: numpy array of string which store
            design variable names and constraint names in snopt
            internal process
        :returns:  pointer to constraint value & gradient,
            array of lower bound constraints,
            array of upper bound constraints,
            design variable names and constraint names,
            n_constraints,
            function computation
        :rtype: numpy array of np.float64,
            numpy array of np.float64,
            numpy array of string,
            integer
        """
        blc = array(())
        buc = array(())
        if self.__n_eq_constraints > 0:
            blc = zeros((self.__n_eq_constraints,))
            buc = zeros((self.__n_eq_constraints,))
            ceqlist = ["c_eq" + str(i) for i in range(self.__n_eq_constraints)]
            names = append(names, array(ceqlist, dtype=str))
        if self.__n_ineq_constraints > 0:
            min_inf = ones(self.__n_ineq_constraints) * -INFINITY
            blc = append(blc, min_inf)
            buc = append(buc, zeros(self.__n_ineq_constraints))
            cieqlist = ["c_ie" + str(i) for i in range(self.__n_ineq_constraints)]
            names = append(names, array(cieqlist, dtype=str))
        # Mandatory dummy free row if unconstrained
        n_constraints = self.__n_ineq_constraints + self.__n_eq_constraints
        if n_constraints == 0:
            n_constraints = 1
            funcon = self.cb_snopt_dummy_func
            blc = append(blc, ones((1,)) * -INFINITY)
            buc = append(buc, ones((1,)) * INFINITY)
            names = append(names, array(["dummy"], dtype=np_str))
        else:
            funcon = self.cb_opt_constraints_snoptb
        return funcon, blc, buc, names, n_constraints

    def _run(self, **options):
        """Runs the algorim, to be overloaded by subclasses.

        :param options: the options dict for the algorithm,
            see associated JSON file
        """
        normalize_ds = options.pop(self.NORMALIZE_DESIGN_SPACE_OPTION, True)
        # Get the  bounds anx x0
        x_0, l_b, u_b = self.get_x0_and_bounds_vects(normalize_ds)
        self.__n_ineq_constraints = self.problem.get_ineq_cstr_total_dim()
        self.__n_eq_constraints = self.problem.get_eq_cstr_total_dim()

        snopt_problem = SNOPT_solver(name="SNOPTB")

        for (key, value) in options.items():
            try:
                snopt_problem.setOption(key.replace("_", " "), value)
            except RuntimeError:
                raise KeyError(
                    "Unknown option or incorrect type :" + str(key) + ":" + str(value)
                )
        snopt_problem.setOption("Verbose", True)
        snopt_problem.setOption("Solution print", True)

        #        n_nonlinear_constraints = n_ineq_constraints + n_eq_constraints
        #        It is assumed that all constraints provided to optimizer are not
        #        inear
        #        n_linear_constraints = 0
        #        n_nonlinear_constraints = n_ineq_constraints + n_eq_constraints

        n_design_variables = x_0.shape[0]

        # Get the normalized bounds:
        l_b = where(isinf(l_b), -INFINITY, l_b)
        u_b = where(isinf(u_b), INFINITY, u_b)

        names = array(["dv" + str(i) for i in range(n_design_variables)], dtype=str)
        i_obj = array([0], np_int)

        funcon, blc, buc, names, n_constraints = self.__preprocess_snopt_constraints(
            names
        )
        n_nl_func = n_constraints
        x0_snopt = append(x_0, zeros((n_constraints,), dtype=float64))
        lower_bounds = concatenate((l_b, blc))
        upper_bounds = concatenate((u_b, buc))
        jacobian = ones((n_nl_func, x_0.shape[0]))
        obj_add = array([0.0], np_float)
        snopt_problem.snoptb(
            # name='SNOPTB',
            m=n_nl_func,
            x0=x0_snopt,
            n=n_design_variables,
            nnCon=n_constraints,
            nnObj=n_design_variables,
            nnJac=n_design_variables,
            iObj=i_obj,
            bl=lower_bounds,
            bu=upper_bounds,
            J=jacobian,
            ObjAdd=obj_add,
            funcon=funcon,
            Names=names,
            funobj=self.cb_opt_objective_snoptb,
        )

        message = self.MESSAGES_DICT[snopt_problem.info]
        status = snopt_problem.info
        return self.get_optimum_from_database(message, status)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Jean-Christophe Giret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Concatenation of several input variables into a single one.

The :class:`.ConcatenationDiscipline` enables to concatenate inputs variables into a
single output variable.

Example:

    This example demonstrates the use of the :class:`.ConcatenationDiscipline`
    instances.
    The contraints variables :math:`c_1` and :math:`c_2` are concatenated into a
    single variable :math:`c`.

        >>> from gemseo.api import create_discipline
        >>> sellar_system_disc = create_discipline('SellarSystem')
        >>> input_vars = ['c1', 'c2']
        >>> output_var = ['c']
        >>> concatenation_disc = create_discipline('ConcatenationDiscipline',
        ...                                         input_vars,
        ...                                         output_var)
        >>> disciplines = [sellar_system_disc, concatenation_disc]
        >>> chain = create_discipline('MDOChain', disciplines=disciplines)
        >>> print(chain.execute())
        >>> print(chain.linearize(force_all=True))
"""
from typing import Optional, Sequence

from numpy import diag, ones, zeros

from gemseo.core.discipline import MDODiscipline
from gemseo.utils.py23_compat import accumulate


class ConcatenationDiscipline(MDODiscipline):
    """Concatenate input variables into a single output variable."""

    def __init__(
        self,
        input_variables,  # type: Sequence[str]
        output_variable,  # type: str
    ):  # type: (...) -> None
        """
        Args:
            input_variables: The input variables to concatenate.
            output_variable: The output variable name.
        """
        super(ConcatenationDiscipline, self).__init__()
        self.input_grammar.initialize_from_data_names(input_variables)
        self.output_grammar.initialize_from_data_names([output_variable])
        self.__output_variable = output_variable

    def _run(self):  # type: (...) -> None
        """Run the discipline."""
        self.local_data[self.__output_variable] = self.get_inputs_asarray()

    def _compute_jacobian(
        self,
        inputs=None,  # type: Optional[Sequence[str]]
        outputs=None,  # type: Optional[Sequence[str]]
    ):  # type: (...) -> None
        """Compute the jacobian matrix.

        Args:
            inputs: The linearization should be performed with respect
                to inputs list. If None, linearization should
                be performed wrt all inputs (Default value = None)
            outputs: The linearization should be performed on outputs list.
                If None, linearization should be performed
                on all outputs (Default value = None)
        """
        self._init_jacobian(inputs, outputs, with_zeros=True)

        inputs_names = self.get_input_data_names()
        inputs_sizes = [inp.size for inp in self.get_all_inputs()]
        inputs_total_size = self.get_inputs_asarray().size

        # Instead of manually accumulating, we use the accumulate() iterator.
        for name, size, start in zip(
            inputs_names, inputs_sizes, accumulate(inputs_sizes, initial=0)
        ):
            val = zeros([inputs_total_size, size])
            val[start : (start + size), :] = diag(ones(size))
            self.jac[self.__output_variable][name] = val



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard, Matthias De Lozzo, Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""The Support Vector Machine algorithm for classification.

This module implements the SVMClassifier class.
A support vector machine (SVM) passes the data through a kernel
in order to increase its dimension
and thereby make the classes linearly separable.

Dependence
----------
The classifier relies on the SVC class
of the `scikit-learn library <https://scikit-learn.org/stable/modules/
generated/sklearn.svm.SVC.html>`_.
"""
from __future__ import division, unicode_literals

import logging
from typing import Callable, Iterable, Optional, Union

from numpy import ndarray
from sklearn.svm import SVC

from gemseo.core.dataset import Dataset
from gemseo.mlearning.classification.classification import MLClassificationAlgo
from gemseo.mlearning.core.ml_algo import TransformerType

LOGGER = logging.getLogger(__name__)


class SVMClassifier(MLClassificationAlgo):
    """The Support Vector Machine algorithm for classification."""

    LIBRARY = "scikit-learn"
    ABBR = "SVM"

    def __init__(
        self,
        data,  # type: Dataset
        transformer=None,  # type: Optional[TransformerType]
        input_names=None,  # type: Optional[Iterable[str]]
        output_names=None,  # type: Optional[Iterable[str]]
        C=1.0,  # noqa: N803 # type: float
        kernel="rbf",  # type: Optional[str,Callable]
        probability=False,  # type: bool
        **parameters  # type: Optional[Union[int,float,bool,str]]
    ):  # type: (...) -> None
        """# noqa: D205,D212,D415
        Args:
            C: The inverse L2 regularization parameter.
                   Higher values give less regularization.
            kernel: The name of the kernel or a callable for the SVM.
                Examples: "linear", "poly", "rbf", "sigmoid", "precomputed"
                or a callable.
            probability: Whether to enable the probability estimates.
                The algorithm is faster if set to False.
        """
        super(SVMClassifier, self).__init__(
            data,
            transformer=transformer,
            input_names=input_names,
            output_names=output_names,
            C=C,
            kernel=kernel,
            probability=probability,
            **parameters
        )
        self.algo = SVC(C=C, kernel=kernel, probability=probability, **parameters)

    def _fit(
        self,
        input_data,  # type:ndarray
        output_data,  # type:ndarray
    ):  # type: (...) -> None
        self.algo.fit(input_data, output_data.ravel())

    def _predict(
        self,
        input_data,  # type:ndarray
    ):  # type: (...) -> ndarray
        return self.algo.predict(input_data)[:, None].astype(int)

    def _predict_proba_soft(
        self,
        input_data,  # type: ndarray
    ):  # type: (...)-> ndarray
        if not self.parameters["probability"]:
            raise NotImplementedError(
                "SVMClassifier soft probability prediction is only available if the "
                "parameter 'probability' is set to True."
            )
        return self.algo.predict_proba(input_data)[..., None]



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""This module contains the base class for machine learning algorithms.

Machine learning is the art of building models from data,
the latter being samples of properties of interest
that can sometimes be sorted by group, such as inputs, outputs, categories, ...

In the absence of such groups,
the data can be analyzed through a study of commonalities,
leading to plausible clusters.
This is referred to as clustering,
a branch of unsupervised learning
dedicated to the detection of patterns in unlabeled data.

.. seealso::

   :mod:`~gemseo.mlearning.core.unsupervised`,
   :mod:`~gemseo.mlearning.cluster.cluster`

When data can be separated into at least two categories by a human,
supervised learning can start with classification
whose purpose is to model the relations
between these categories and the properties of interest.
Once trained,
a classification model can predict the category
corresponding to new property values.

.. seealso::

   :mod:`~gemseo.mlearning.core.supervised`,
   :mod:`~gemseo.mlearning.classification.classification`

When the distinction between inputs and outputs can be made among the data properties,
another branch of supervised learning can be considered: regression modeling.
Once trained,
a regression model can predict the outputs corresponding to new inputs values.

.. seealso::

   :mod:`~gemseo.mlearning.core.supervised`,
   :mod:`~gemseo.mlearning.regression.regression`

The quality of a machine learning algorithm can be measured
using a :class:`.MLQualityMeasure`
either with respect to the learning dataset
or to a test dataset or using resampling methods,
such as K-folds or leave-one-out cross-validation techniques.
The challenge is to avoid over-learning the learning data
leading to a loss of generality.
We often want to build models that are not too dataset-dependent.
For that,
we want to maximize both a learning quality and a generalization quality.
In unsupervised learning,
a quality measure can represent the robustness of clusters definition
while in supervised learning, a quality measure can be interpreted as an error,
whether it is a misclassification in the case of the classification algorithms
or a prediction one in the case of the regression algorithms.
This quality can often be improved
by building machine learning models from standardized data
in such a way that the data properties have the same order of magnitude.


.. seealso::

   :mod:`~gemseo.mlearning.qual_measure.quality_measure`,
   :mod:`~gemseo.mlearning.transform.transformer`

Lastly,
a machine learning algorithm often depends on hyperparameters
to be carefully tuned in order to maximize the generalization power of the model.

.. seealso::

   :mod:`~gemseo.mlearning.core.calibration`
   :mod:`~gemseo.mlearning.core.selection`
"""
from __future__ import division, unicode_literals

import logging
import pickle
import re
from os import makedirs
from os.path import exists, join
from typing import Any, Dict, List, Optional, Union

import six
from custom_inherit import DocInheritMeta
from numpy import ndarray

from gemseo.core.dataset import Dataset
from gemseo.mlearning.transform.transformer import Transformer
from gemseo.utils.string_tools import MultiLineString, pretty_repr

LOGGER = logging.getLogger(__name__)

TransformerType = Dict[str, Transformer]
SavedObjectType = Union[Dataset, TransformerType, str, bool]
DataType = Union[ndarray, Dict[str, ndarray]]
MLAlgoParameterType = Optional[Any]


@six.add_metaclass(
    DocInheritMeta(
        abstract_base_class=True,
        style="google_with_merge",
    )
)
class MLAlgo(object):
    """An abstract machine learning algorithm.

    Such a model is built from a training dataset,
    data transformation options and parameters. This abstract class defines the
    :meth:`.MLAlgo.learn`, :meth:`.MLAlgo.save` methods and the boolean
    property, :attr:`!MLAlgo.is_trained`. It also offers a string
    representation for end users.
    Derived classes shall overload the :meth:`.MLAlgo.learn`,
    :meth:`!MLAlgo._save_algo` and :meth:`!MLAlgo._load_algo` methods.

    Attributes:
        learning_set (Dataset): The learning dataset.
        parameters (Dict[str,MLAlgoParameterType]): The parameters
            of the machine learning algorithm.
        transformer (Dict[str,Transformer]): The strategies to transform the variables.
            The values are instances of :class:`.Transformer`
            while the keys are the names of
            either the variables
            or the groups of variables,
            e.g. "inputs" or "outputs" in the case of the regression algorithms.
            If a group is specified,
            the :class:`.Transformer` will be applied
            to all the variables of this group.
            If None, do not transform the variables.
        algo (Any): The interfaced machine learning algorithm.
    """

    LIBRARY = None
    ABBR = "MLAlgo"
    FILENAME = "ml_algo.pkl"

    def __init__(
        self,
        data,  # type:Dataset
        transformer=None,  # type: Optional[TransformerType]
        **parameters  # type: MLAlgoParameterType
    ):  # type: (...) -> None
        """
        Args:
            data: The learning dataset.
            transformer: The strategies to transform the variables.
                The values are instances of :class:`.Transformer`
                while the keys are the names of
                either the variables
                or the groups of variables,
                e.g. "inputs" or "outputs" in the case of the regression algorithms.
                If a group is specified,
                the :class:`.Transformer` will be applied
                to all the variables of this group.
                If None, do not transform the variables.
            **parameters: The parameters of the machine learning algorithm.
        """
        self.learning_set = data
        self.parameters = parameters
        self.transformer = {}
        if transformer:
            self.transformer = {
                group: transf.duplicate() for group, transf in transformer.items()
            }

        self.algo = None
        self._trained = False

    class DataFormatters(object):
        """Decorators for the internal MLAlgo methods."""

    @property
    def is_trained(self):  # type: (...) -> bool
        """Return whether the algorithm is trained."""
        return self._trained

    def learn(
        self,
        samples=None,  # type: List[int]
    ):  # type: (...) -> None
        """Train the machine learning algorithm from the learning dataset.

        Args:
            samples: The indices of the learning samples.
                If None, use the whole learning dataset.
        """
        raise NotImplementedError

    def __str__(self):  # type: (...) -> str
        msg = MultiLineString()
        msg.add("{}({})", self.__class__.__name__, pretty_repr(self.parameters))
        msg.indent()
        if self.LIBRARY is not None:
            msg.add("based on the {} library", self.LIBRARY)
        msg.add("built from {} learning samples", self.learning_set.length)
        return str(msg)

    def save(
        self,
        directory=None,  # type: Optional[str]
        path=".",  # type: str
        save_learning_set=False,  # type: bool
    ):  # type: (...) -> str
        """Save the machine learning algorithm.

        Args:
            directory: The name of the directory to save the algorithm.
            path: The path to parent directory where to create the directory.
            save_learning_set: If False, do not save the learning set
                to lighten the saved files.

        Returns:
            The path to the directory where the algorithm is saved.
        """
        if not save_learning_set:
            self.learning_set.data = {}
            self.learning_set.length = 0
        splitted_class_name = re.findall("[A-Z][a-z]*", self.__class__.__name__)
        splitted_class_name = [word.lower() for word in splitted_class_name]
        algo_name = "_".join(splitted_class_name)
        algo_name += "_" + self.learning_set.name
        directory = directory or algo_name
        directory = join(path, directory)
        if not exists(directory):
            makedirs(directory)

        filename = join(directory, self.FILENAME)
        objects = self._get_objects_to_save()
        with open(filename, "wb") as handle:
            pickle.dump(objects, handle)

        self._save_algo(directory)

        return directory

    def _save_algo(
        self,
        directory,  # type: str
    ):  # type: (...) -> None
        """Save the interfaced machine learning algorithm.

        Args:
            directory: The path to the directory
                where to save the interfaced machine learning algorithm.
        """
        filename = join(directory, "algo.pkl")
        with open(filename, "wb") as handle:
            pickle.dump(self.algo, handle)

    def load_algo(
        self,
        directory,  # type: str
    ):  # type: (...) -> None
        """Load a machine learning algorithm from a directory.

        Args:
            directory: The path to the directory
                where the machine learning algorithm is saved.
        """

        filename = join(directory, "algo.pkl")
        with open(filename, "rb") as handle:
            algo = pickle.load(handle)
        self.algo = algo

    def _get_objects_to_save(self):  # type: (...) -> Dict[str,SavedObjectType]
        """Return the objects to save.

        Returns:
            The objects to save.
        """
        objects = {
            "data": self.learning_set,
            "transformer": self.transformer,
            "parameters": self.parameters,
            "algo_name": self.__class__.__name__,
            "_trained": self._trained,
        }
        return objects



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Damien Guenot
#        :author: Francois Gallard, refactoring
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Optimization library wrappers base class."""
from __future__ import division, unicode_literals

import logging

from gemseo.algos.driver_lib import DriverLib
from gemseo.algos.stop_criteria import (
    FtolReached,
    XtolReached,
    is_f_tol_reached,
    is_x_tol_reached,
)

LOGGER = logging.getLogger(__name__)


class OptimizationLibrary(DriverLib):
    """Optimization library base class See DriverLib."""

    MAX_ITER = "max_iter"
    F_TOL_REL = "ftol_rel"
    F_TOL_ABS = "ftol_abs"
    X_TOL_REL = "xtol_rel"
    X_TOL_ABS = "xtol_abs"
    STOP_CRIT_NX = "stop_crit_n_x"
    # Maximum step for the line search
    LS_STEP_SIZE_MAX = "max_ls_step_size"
    # Maximum number of line search steps (per iteration).
    LS_STEP_NB_MAX = "max_ls_step_nb"
    MAX_FUN_EVAL = "max_fun_eval"
    MAX_TIME = "max_time"
    PG_TOL = "pg_tol"
    VERBOSE = "verbose"

    def __init__(self):
        super(OptimizationLibrary, self).__init__()
        self._ftol_rel = 0.0
        self._ftol_abs = 0.0
        self._xtol_rel = 0.0
        self._xtol_abs = 0.0
        self._stop_crit_n_x = 3

    def __algorithm_handles(self, algo_name, cstr_type):
        """Returns True if the algorithms handles cstr_type constraints.

        :param algo_name : the name of the algo_name
        :returns: True or False
        """
        if algo_name not in self.lib_dict:
            raise KeyError(
                "Algorithm "
                + str(algo_name)
                + " not in library "
                + self.__class__.__name__
            )
        if cstr_type in self.lib_dict[algo_name]:
            return self.lib_dict[algo_name][cstr_type]
        return False

    def algorithm_handles_eqcstr(self, algo_name):
        """Returns True if the algorithms handles equality constraints.

        :param algo_name: the name of the algorithm
        :returns: True or False
        """
        return self.__algorithm_handles(algo_name, self.HANDLE_EQ_CONS)

    def algorithm_handles_ineqcstr(self, algo_name):
        """Returns True if the algorithms handles inequality constraints.

        :param algo_name: the name of the algorithm
        :returns: True or False
        """
        return self.__algorithm_handles(algo_name, self.HANDLE_INEQ_CONS)

    def is_algo_requires_positive_cstr(self, algo_name):
        """Returns True if the algorithm requires positive constraints False otherwise.

        :param algo_name: the name of the algorithm
        :returns: True if constraints must be positive
        :rtype: logical
        """
        loc_dict = self.lib_dict[algo_name]
        if self.POSITIVE_CONSTRAINTS in loc_dict:
            return loc_dict[self.POSITIVE_CONSTRAINTS]
        return False

    def _check_constraints_handling(self, algo_name, problem):
        """Check if problem and algorithm are consistent for constraints handling."""
        if problem.has_eq_constraints() and not self.algorithm_handles_eqcstr(
            algo_name
        ):
            raise ValueError(
                "Requested optimization algorithm "
                + "%s can not handle equality constraints" % self.algo_name
            )
        if problem.has_ineq_constraints() and not self.algorithm_handles_ineqcstr(
            algo_name
        ):
            raise ValueError(
                "Requested optimization algorithm "
                + " %s can not handle inequality constraints" % algo_name
            )

    def get_right_sign_constraints(self):
        """Transforms the problem constraints into their opposite sign counterpart if
        the algorithm requires positive constraints."""
        if self.problem.has_ineq_constraints() and self.is_algo_requires_positive_cstr(
            self.algo_name
        ):
            return [-cstr for cstr in self.problem.constraints]
        return self.problem.constraints

    def _run(self, **options):
        """Runs the algorithm, to be overloaded by subclasses.

        :param options: the options dict for the algorithm,
            see associated JSON file
        """
        raise NotImplementedError()

    def _pre_run(self, problem, algo_name, **options):
        """To be overriden by subclasses Specific method to be executed just before _run
        method call.

        :param problem: the problem to be solved
        :param algo_name: name of the algorithm
        :param options: the options dict for the algorithm,
            see associated JSON file
        """
        super(OptimizationLibrary, self)._pre_run(problem, algo_name, **options)
        self._check_constraints_handling(algo_name, problem)

        if self.MAX_ITER in options:
            max_iter = options[self.MAX_ITER]
        elif (
            self.MAX_ITER in self.OPTIONS_MAP
            and self.OPTIONS_MAP[self.MAX_ITER] in options
        ):
            max_iter = options[self.OPTIONS_MAP[self.MAX_ITER]]
        else:
            raise ValueError("Could not determine the maximum number of iterations")

        self._ftol_rel = options.get(self.F_TOL_REL, 0.0)
        self._ftol_abs = options.get(self.F_TOL_ABS, 0.0)
        self._xtol_rel = options.get(self.X_TOL_REL, 0.0)
        self._xtol_abs = options.get(self.X_TOL_ABS, 0.0)
        self._stop_crit_n_x = options.get(self.STOP_CRIT_NX, 3)
        LOGGER.info("%s", problem)
        if problem.design_space.dimension <= self.MAX_DS_SIZE_PRINT:
            LOGGER.info("%s", problem.design_space)
        self.init_iter_observer(max_iter, "Optimization")
        problem.add_callback(self.new_iteration_callback)
        eval_jac = self.is_algo_requires_grad(algo_name)
        normalize = options.get(self.NORMALIZE_DESIGN_SPACE_OPTION, True)
        # First, evaluate all functions at x_0. Some algorithms dont do this
        self.problem.evaluate_functions(
            eval_jac=eval_jac, eval_obj=True, normalize=normalize
        )

    @staticmethod
    def is_algorithm_suited(algo_dict, problem):
        """Checks if the algorithm is suited to the problem according to its algo dict.

        :param algo_dict: the algorithm characteristics
        :param problem: the opt_problem to be solved
        """
        if problem.has_eq_constraints() and (
            (OptimizationLibrary.HANDLE_EQ_CONS not in algo_dict)
            or not algo_dict[OptimizationLibrary.HANDLE_EQ_CONS]
        ):
            return False
        if problem.has_ineq_constraints() and (
            (OptimizationLibrary.HANDLE_INEQ_CONS not in algo_dict)
            or not algo_dict[OptimizationLibrary.HANDLE_INEQ_CONS]
        ):
            return False
        non_lin = problem.pb_type == problem.NON_LINEAR_PB
        lin_alg = (
            OptimizationLibrary.PROBLEM_TYPE in algo_dict
            and algo_dict[OptimizationLibrary.PROBLEM_TYPE] == problem.LINEAR_PB
        )
        if non_lin and lin_alg:
            return False

        return True

    def new_iteration_callback(self):
        """Callback called at each new iteration, ie every time a design vector that is
        not already in the database is proposed by the optimizer.

        Iterates the progress bar, implements the stop criteria
        """
        # First check if the max_iter is reached and update the progress bar
        super(OptimizationLibrary, self).new_iteration_callback()
        if is_f_tol_reached(
            self.problem, self._ftol_rel, self._ftol_abs, self._stop_crit_n_x
        ):
            raise FtolReached()
        if is_x_tol_reached(
            self.problem, self._xtol_rel, self._xtol_abs, self._stop_crit_n_x
        ):
            raise XtolReached()



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# -*-mode: python; py-indent-offset: 4; tab-width: 8; coding:utf-8 -*-
# Copyright (c) 2018 IRT-AESE.
# All rights reserved.
#
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Arthur Piat
#        :author: Nicolas Roussouly: GEMSEO integration

"""Definition of the Matlab license manager.

Overview
--------

This module contains the :class:`.LicenseManager`
which enables to check the presence of any toolbox licenses
of the Matlab installation.
"""

import logging
from typing import Optional

from gemseo.utils.py23_compat import Path
from gemseo.wrappers.matlab.engine.engine import __MatlabEngine

LOGGER = logging.getLogger(__name__)


class LicenseManager(object):
    """Manage Matlab License.

    The licenseManager was created to enables de-synchronised
    launch of optimization using matlab_discipline. The goal
    is to wait until all licenses that are needed are available
    in Matlab workspace.
    Parallel computing launch can be used with this tool.

    Attributes:
        engine: The MatlabEngine instance.
        licenses: The list of license toolbox that must be used.

    Examples:
        >>> # Build a new matlab engine
        >>> eng = get_matlab_engine()
        >>> # add a toolbox to the engine
        >>> eng.add_toolbox("signal_toolbox")
        >>> # build a license manager from the previous engine
        >>> lm = LicenseManager(eng)
        >>> # check licenses of the engine until all are available
        >>> lm.check_licenses()
    """

    SIGNAL_TOOL = "signal_toolbox"
    DISTRIB_COMP_TOOL = "distrib_computing_toolbox"
    CURVE_FIT_TOOL = "Curve_Fitting_Toolbox"

    def __init__(
        self, engine  # type: __MatlabEngine
    ):  # type: (...) -> None
        """# noqa: D205,D212,D415
        Args:
            engine: The MatlabEngine instance.
        """
        self.__engine = engine
        if self.__engine.is_closed:
            # If engine is not (re)started here, add_path
            # will raise an error. A closed engine could happen here
            # since engine is a singleton and could be closed in any
            # other location
            self.__engine.start_engine()

        self.__engine.add_path(str(Path(__file__).parent / "matlab_files"))
        self.__licenses = self.__engine.get_toolboxes()

    @property
    def licenses(self):
        """Get and set all licences."""
        return self.__licenses

    @licenses.setter
    def licenses(self, licenses):
        self.__licenses = licenses

    def check_licenses(
        self,
        licenses=None,  # type: Optional[list[str]]
        pause_frac=60,  # type: float
        pause_const=20,  # type: float
    ):  # type: (...) -> None
        """Check that the Matlab licenses exist.

        The method fetches all the needed licenses thanks to the
        matlab function licenseControl and the class Logger. Note
        that the MATLAB function will be looping until the given
        toolboxes are available.

        Args:
            licenses: The list of matlab toolboxes.
                If ``None``, use the already existing engine licenses.
            pause_frac: The time used between each try to get licenses.
            pause_const: The time added in order to estimate the waiting time.
                The waiting time is estimated at each try with the following formula:
                ``Wt = pause_const + random([0,1])*pause_frac``.
        """
        if licenses is not None:
            self.licenses = licenses

        if self.licenses is None or not self.licenses:
            LOGGER.info("No MATLAB license check will be performed.")
            return

        self.__engine.execute_function(
            "licenseControl", self.licenses, pause_frac, pause_const, nargout=0
        )

    def start_parallel_computing(
        self,
        n_parallel_workers=4,  # type: int
        cluster_name="local",  # type: str
    ):  # type: (...) -> bool
        """Start parallel computing in MatlabEngine.

        Args:
            n_parallel_workers: The number of "workers" to the parallel pool.
                Maximum number allowed is 12
            cluster_name: The matlab parallel pool cluster name. If None,
                the given name is the same as the engine.
        """
        self.check_licenses(self.DISTRIB_COMP_TOOL)
        self.__engine.start_parallel_computing(n_parallel_workers, cluster_name)

    def end_parallel_computing(self):  # type: (...) -> bool
        """Stop parallel computing in matlab engine."""
        self.__engine.end_parallel_computing()
